#!/usr/bin/env python
#
# floatspin.py - Alternate implementation to wx.SpinCtrlDouble and
#                wx.lib.agw.floatspin.FloatSpin.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`FloatSpinCtrl` class, a spin control for
modifying a floating point value.
"""


import time
import logging

import                    wx
import wx.lib.newevent as wxevent

import fsleyes_widgets as fw


log = logging.getLogger(__name__)


if fw.wxversion() == fw.WX_PHOENIX: FloatSpinBase = wx.Panel
else:                               FloatSpinBase = wx.PyPanel


class FloatSpinCtrl(FloatSpinBase):
    """A ``FloatSpinCtrl`` is a :class:`wx.Panel` which contains a
    :class:`wx.TextCtrl` and a :class:`wx.SpinButton`, allowing the user to
    modify a floating point (or integer) value.

    The ``FloatSpinCtrl`` is an alternative to :class:`wx.SpinCtrl`,
    :class:`wx.SpinCtrlDouble`, and :class:`wx.lib.agw.floatspin.FloatSpin`.

     - :class:`wx.SpinCtrlDouble`: Under Linux/GTK, this widget does not allow
        the user to enter values that are not a multiple of the increment.

     - :class:`wx.lib.agw.floatspin.FloatSpin`. Differs from the
       :class:`wx.SpinCtrl` API in various annoying ways, and formatting is a
       pain.
    """

    def __init__(self,
                 parent,
                 minValue=None,
                 maxValue=None,
                 increment=None,
                 value=None,
                 style=None,
                 width=None,
                 evDelta=None):
        """Create a ``FloatSpinCtrl``.

        The following style flags are available:
          .. autosummary::
             FSC_MOUSEWHEEL
             FSC_INTEGER
             FSC_NO_LIMIT

        :arg parent:    The parent of this control (e.g. a :class:`wx.Panel`).

        :arg minValue:  Initial minimum value.

        :arg maxValue:  Initial maximum value.

        :arg increment: Default increment to apply when the user changes the
                        value via the spin button or mouse wheel.

        :arg value:     Initial value.

        :arg style:     Style flags - a combination of :data:`FSC_MOUSEWHEEL`,
                        :data:`FSC_INTEGER`, and :data:`FSC_NO_LIMIT`.

        :arg width:     If provided, desired text control width (in
                        characters).

        :arg evDelta:   Minimum time between consecutive ``wx.SpinButton``
                        events. On Linux/GTK, the wx.SpinButton is badly
                        behaved - if, while clicking on the mouse button, the
                        user moves the mouse even a tiny bit, more than one
                        spin event will be generated. To work around this
                        (without having to write my own ``wx.SpinButton``
                        implementation), the ``evDelta`` parameter allows me
                        to throttle the maximum rate at which events received
                        from the spin button can be processed. This is
                        implemented in the :meth:`__onSpinDown` and
                        :meth:`__onSpinUp` methods.

                        This has the side effect that if the user clicks and
                        holds on the spin button, they have to wait <delta>
                        seconds between increments/decrements.
        """
        FloatSpinBase.__init__(self, parent)

        if minValue  is None: minValue  = 0
        if maxValue  is None: maxValue  = 100
        if value     is None: value     = 0
        if increment is None: increment = 1
        if style     is None: style     = 0
        if evDelta   is None: evDelta   = 0.5

        self.__integer = style & FSC_INTEGER
        self.__nolimit = style & FSC_NO_LIMIT

        self.__value     = value
        self.__increment = increment
        self.__realMin   = float(minValue)
        self.__realMax   = float(maxValue)
        self.__realRange = abs(self.__realMax - self.__realMin)

        # Attributes used in spin
        # button event rate throttling
        self.__lastEvent  = time.time()
        self.__eventDelta = evDelta

        # We use the full signed 32 bit integer
        # range offered by the wx.SpinButton class.
        self.__realSpinMin = -2 ** 31
        self.__realSpinMax =  2 ** 31 - 1

        # Unless the no limit style has been
        # specified, in which case we map the
        # real data range to 16 bits, and
        # allow the rest of the 32 bit range
        # to account for overflow. In either
        # case, the spin button is configured
        # to use the full 32 bit range.
        if  self.__nolimit:
            self.__spinMin = -2 ** 15
            self.__spinMax =  2 ** 15 - 1
        else:
            self.__spinMin = self.__realSpinMin
            self.__spinMax = self.__realSpinMax

        self.__spinRange = abs(self.__spinMax - self.__spinMin)

        self.__text = wx.TextCtrl(  self,
                                    style=wx.TE_PROCESS_ENTER)
        self.__spin = wx.SpinButton(self,
                                    style=wx.SP_VERTICAL | wx.SP_ARROW_KEYS)

        self.__spin.SetRange(self.__realSpinMin, self.__realSpinMax)

        if width is not None:
            width = self.__text.GetTextExtent('0' * width)[0]
            self.__text.SetMinSize((width + 10, -1))
            self.__text.SetMaxSize((width + 10, -1))

        if self.__integer: self.__format = '{:d}'
        else:              self.__format = '{:.7G}'

        # Events on key down, enter, focus
        # lost, and on the spin control
        self.__text.Bind(wx.EVT_KEY_DOWN,   self.__onKeyDown)
        self.__text.Bind(wx.EVT_TEXT_ENTER, self.__onText)
        self.__text.Bind(wx.EVT_KILL_FOCUS, self.__onKillFocus)
        self.__spin.Bind(wx.EVT_SPIN_UP,    self.__onSpinUp)
        self.__spin.Bind(wx.EVT_SPIN_DOWN,  self.__onSpinDown)

        # Event on mousewheel
        # if style enabled
        if style & FSC_MOUSEWHEEL:
            self.__spin.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)
            self.__text.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)

        # Under linux/GTK, text controls absorb
        # mousewheel events, so we bind our own
        # handler to prevent this.
        elif wx.Platform == '__WXGTK__':
            def wheel(ev):
                self.GetParent().GetEventHandler().ProcessEvent(ev)
            self.__spin.Bind(wx.EVT_MOUSEWHEEL, wheel)
            self.__text.Bind(wx.EVT_MOUSEWHEEL, wheel)

        # Under linux/GTK, double-clicking the
        # textctrl selects the word underneath
        # the cursor, whereas we want it to
        # select the entire textctrl contents.
        # Mouse event behaviour cannot be overridden
        # under OSX, but its behaviour is more
        # sensible, so a hack is not necessary.
        if wx.Platform == '__WXGTK__':
            self.__text.Bind(wx.EVT_LEFT_DCLICK, self.__onDoubleClick)

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)

        self.__sizer.Add(self.__text, flag=wx.EXPAND, proportion=1)
        self.__sizer.Add(self.__spin)

        self.Layout()

        self.SetSizer(  self.__sizer)
        self.SetMinSize(self.__sizer.GetMinSize())

        self.SetRange(minValue, maxValue)
        self.SetIncrement(increment)


    @property
    def textCtrl(self):
        """Returns a reference to the ``TextCtrl`` contained in this
        ``FloatSpinCtrl``.
        """
        return self.__text


    @property
    def spinButton(self):
        """Returns a reference to the ``SpinButton`` contained in this
        ``FloatSpinCtrl``.
        """
        return self.__spin


    def DoGetBestClientSize(self):
        """Returns the best size for this ``FloatSpinCtrl``.
        """
        return self.__sizer.GetMinSize()


    def GetValue(self):
        """Returns the current value."""
        return self.__value


    def GetMin(self):
        """Returns the current minimum value."""
        return float(self.__realMin)


    def GetMax(self):
        """Returns the current maximum value."""
        return float(self.__realMax)


    def GetIncrement(self):
        """Returns the current inrement."""
        return self.__increment


    def SetIncrement(self, inc):
        """Sets the inrement."""
        if self.__integer: self.__increment = int(round(inc))
        else:              self.__increment =           inc


    def GetRange(self):
        """Returns the current data range, a tuple containing the
        ``(min, max)`` values.
        """
        return (self.__realMin, self.__realMax)


    def SetMin(self, minval):
        """Sets the minimum value."""
        self.SetRange(minval, self.__realMax)


    def SetMax(self, maxval):
        """Sets the maximum value."""
        self.SetRange(self.__realMin, maxval)


    def SetRange(self, minval, maxval):
        """Sets the minimum and maximum values."""

        if minval > maxval:
            raise ValueError('Min cannot be greater than max '
                             '({} > {})'.format(minval, maxval))

        if self.__integer:
            minval = int(round(minval))
            maxval = int(round(maxval))

        self.__realMin   = float(minval)
        self.__realMax   = float(maxval)
        self.__realRange = abs(self.__realMax - self.__realMin)

        self.SetValue(self.__value)


    def SetValue(self, newValue):
        """Sets the value.

        :returns ``True`` if the value was changed, ``False`` otherwise.
        """

        # Clamp the value so it stays
        # within the min/max, unless the
        # FSC_NO_LIMIT style flag is set.
        if not self.__nolimit:
            if newValue < self.__realMin: newValue = self.__realMin
            if newValue > self.__realMax: newValue = self.__realMax

        if self.__integer:
            newValue = int(round(newValue))

        oldValue     = self.__value
        self.__value = newValue

        self.__text.ChangeValue(self.__format.format(newValue))

        # The wx.SpinButton is badly behaved. It doesn't have
        # a ChangeValue method (which would explicitly allow
        # us to update the stored spin button value without
        # triggering an event), and under GTK, when the
        # SetValue method is called from a SPIN_UP/SPIN_DOWN
        # event, it will trigger another event. So we disable
        # events from the spin button when setting the value.
        self.__spin.SetEvtHandlerEnabled(False)
        self.__spin.SetValue(self.__realToSpin(newValue))

        # We have to re-enable event processing
        # asynchronously, otherwise the SpinButton
        # will keep generating events.  We check
        # the state of this control before doing
        # so, as the control may get deleted before
        # this function gets called.
        def reset():
            if self and self.__spin:
                self.__spin.SetEvtHandlerEnabled(True)

        wx.CallAfter(reset)

        return newValue != oldValue


    def __onKillFocus(self, ev):
        """Called when the text field of this ``FloatSpinCtrl`` loses focus.
        Generates an :attr:`.EVT_FLOATSPIN` event.
        """
        ev.Skip()

        log.debug('Spin lost focus - simulating text event')
        self.__onText(ev)


    def __onKeyDown(self, ev):
        """Called on ``wx.EVT_KEY_DOWN`` events. If the user pushes the up or
        down arrow keys, the value is changed (using the :meth:`__onSpinUp`
        and :meth:`__onSpinDown` methods).
        """
        up   = wx.WXK_UP
        down = wx.WXK_DOWN
        key  = ev.GetKeyCode()

        log.debug('Key down event: {} (looking for up [{}] '
                  'or down [{}])'.format(key, up, down))

        if   key == up:   self.__onSpinUp()
        elif key == down: self.__onSpinDown()
        else:             ev.Skip()


    def __onText(self, ev):
        """Called when the user changes the value via the text control.

        This method is called when the enter key is pressed.

        If the entered value is a valid number, a ``wx.EVT_TEXT_ENTER``
        event is generated. This event will have a boolean attribute,
        ``changed``, which is ``True`` if the value that was stored was
        different to that entered by the user (e.g. if it was clamped
        to the min/max bounds).

        If the value was changed from its previous value, a
        :data:`FloatSpinEvent` is also generated.
        """

        val = self.__text.GetValue().strip()

        log.debug('Spin text - attempting to change value '
                  'from {} to {}'.format(self.__value, val))

        try:
            if self.__integer: val = int(  val)
            else:              val = float(val)
        except ValueError:
            self.SetValue(self.__value)
            return

        valset = self.SetValue(val)

        # Add a 'changed' attribute so
        # users can tell if the value
        # that was entered was not the
        # value that ended up getting
        # stored
        ev         = wx.PyCommandEvent(wx.EVT_TEXT_ENTER.typeId, self.GetId())
        ev.changed = val != self.__value
        wx.PostEvent(self.GetEventHandler(), ev)

        # Emit a spin event if the value
        # changed from its previous value
        if valset:
            wx.PostEvent(self, FloatSpinEvent(value=self.__value))


    def __onSpinDown(self, ev=None):
        """Called when the *down* button on the ``wx.SpinButton`` is pushed.

        Decrements the value by the current increment and generates a
        :data:`FloatSpinEvent`.
        """

        # See comments in __init__
        if ev is not None:

            lastEv = self.__lastEvent
            thisEv = time.time()

            if thisEv - lastEv < self.__eventDelta:
                return

            self.__lastEvent = thisEv

        log.debug('Spin down button - attempting to change value '
                  'from {} to {}'.format(self.__value,
                                         self.__value - self.__increment))

        if self.SetValue(self.__value - self.__increment):
            wx.PostEvent(self, FloatSpinEvent(value=self.__value))


    def __onSpinUp(self, ev=None):
        """Called when the *up* button on the ``wx.SpinButton`` is pushed.

        Increments the value by the current increment and generates a
        :data:`FloatSpinEvent`.
        """

        # See comments in __init__
        if ev is not None:
            lastEv = self.__lastEvent
            thisEv = time.time()

            if thisEv - lastEv < self.__eventDelta:
                return
            self.__lastEvent = thisEv

        log.debug('Spin up button - attempting to change value '
                  'from {} to {}'.format(self.__value,
                                         self.__value + self.__increment))

        if self.SetValue(self.__value + self.__increment):
            wx.PostEvent(self, FloatSpinEvent(value=self.__value))


    def __onMouseWheel(self, ev):
        """If the :data:`FSC_MOUSEWHEEL` style flag is set, this method is
        called on mouse wheel events.

        Calls :meth:`__onSpinUp` on an upwards rotation, and
        :meth:`__onSpinDown` on a downwards rotation.
        """

        log.debug('Mouse wheel - delegating to spin event handlers')

        rot = ev.GetWheelRotation()

        if ev.GetWheelAxis() == wx.MOUSE_WHEEL_HORIZONTAL:
            rot = -rot

        if   rot > 0: self.__onSpinUp()
        elif rot < 0: self.__onSpinDown()


    def __onDoubleClick(self, ev):
        """Called when the user double clicks in the ``TextCtrl``. Selects
        the entire contents of the ``TextCtrl``.
        """
        self.__text.SelectAll()


    def __realToSpin(self, value):
        """Converts the given value from real space to spin button space."""

        if self.__realRange == 0:
            return 0

        if self.__integer:
            value = int(round(value))

        value     = float(value)
        spinMin   = float(self.__spinMin)
        realMin   = float(self.__realMin)
        realRange = float(self.__realRange)
        spinRange = float(self.__spinRange)

        value     = spinMin + (value - realMin) * (spinRange / realRange)

        # Don't allow the value to flow over
        # the real wx.SpinButton range.
        if   value < self.__realSpinMin: value = self.__realSpinMin
        elif value > self.__realSpinMax: value = self.__realSpinMax

        return int(round(value))


_FloatSpinEvent,      _EVT_FLOATSPIN       = wxevent.NewEvent()
_FloatSpinEnterEvent, _EVT_FLOATSPIN_ENTER = wxevent.NewEvent()


EVT_FLOATSPIN = _EVT_FLOATSPIN
"""Identifier for the :data:`FloatSpinEvent` event. """


FloatSpinEvent = _FloatSpinEvent
"""Event emitted when the floating point value is changed by the user. A
``FloatSpinEvent`` has the following attributes:

  - ``value``: The new value.
"""


FSC_MOUSEWHEEL = 1
"""If set, mouse wheel events on the control will change the value. """


FSC_INTEGER = 2
"""If set, the control stores an integer value, rather than a floating point
value.
"""

FSC_NO_LIMIT = 4
"""If set, the control will allow the user to enter values that are outside
of the current minimum/maximum limits.
"""



#!/usr/bin/env python
#
# __init__.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

from __future__ import print_function

import gc
import time
import traceback

import numpy as np

import wx


GTK = any(['gtk' in p.lower() for p in wx.PlatformInfo])


def compare_images(arr1, arr2, threshold):
    """Compares two images using the normalised difference of arr1 with respect
    to arr2. Returns a tuple containing:

     - A boolean value indicating whether the test passed (the difference was
       less than or equal to the threshold).

     - The normalised difference between the two..
    """
    # Discard alpha values
    arr1 = arr1[:, :, :3]
    arr2 = arr2[:, :, :3]

    if arr1.shape != arr2.shape:
        return False, 0

    finite1 = np.isfinite(arr1)
    finite2 = np.isfinite(arr2)

    if not np.all(finite1 == finite2):
        return 1

    a2zero = arr2 == 0
    nzarr1 = arr1[finite1 & ~a2zero]
    nzarr2 = arr2[finite1 & ~a2zero]
    zarr1  = arr1[finite1 &  a2zero]
    zarr2  = arr2[finite1 &  a2zero]

    if nzarr2.size > 0: zdenom = abs(nzarr2.mean())
    else:               zdenom = 1

    normdiff = np.zeros(arr1.shape)
    normdiff[finite1 & ~a2zero] = np.abs((nzarr2 - nzarr1) / nzarr2)
    normdiff[finite1 &  a2zero] = np.abs((zarr2  - zarr1)  / zdenom)

    if normdiff.size > 0:
        result = normdiff.mean()
        return result <= threshold, result
    else:
        return True, 0


def run_with_wx(func, *args, **kwargs):

    gc.collect()

    propagateRaise = kwargs.pop('propagateRaise', True)
    startingDelay  = kwargs.pop('startingDelay',  500)
    finishingDelay = kwargs.pop('finishingDelay', 500)
    callAfterApp   = kwargs.pop('callAfterApp',   None)

    result = [None]
    raised = [None]

    app    = [wx.App()]
    frame  = [wx.Frame(None)]

    if callAfterApp is not None:
        callAfterApp()

    def wrap():

        try:
            if func is not None:
                result[0] = func(*args, **kwargs)

        except Exception as e:
            traceback.print_exc()
            raised[0] = e

        finally:
            def finish():
                frame[0].Destroy()
                app[0].ExitMainLoop()
            wx.CallLater(finishingDelay, finish)

    frame[0].Show()

    wx.CallLater(startingDelay, wrap)

    app[0].MainLoop()

    time.sleep(1)

    if raised[0] and propagateRaise:
        raise raised[0]

    del app[0]
    del frame[0]

    return result[0]


def addall(parent, widgets):

    sizer = wx.BoxSizer(wx.VERTICAL)
    for w in widgets:
        sizer.Add(w, flag=wx.EXPAND, proportion=1)
    parent.SetSizer(sizer)
    parent.Layout()
    parent.Refresh()
    parent.Update()
    realYield()


# Under GTK, a single call to
# yield just doesn't cut it
def realYield(centis=10):
    for i in range(int(centis)):
        wx.YieldIfNeeded()
        time.sleep(0.01)


# stype:
#   0 for single click
#   1 for double click
#   2 for separatemouse down/up events
def simclick(sim, target, btn=wx.MOUSE_BTN_LEFT, pos=None, stype=0):

    class FakeEv(object):
        def __init__(self, evo):
            self.evo = evo
        def GetEventObject(self):
            return self.evo

    parent = target.GetParent()
    if GTK:

        if type(target).__name__ == 'StaticTextTag' and \
           type(parent).__name__ == 'TextTagPanel':
            parent._TextTagPanel__onTagLeftDown(FakeEv(target))
            realYield()
            return

        if type(target).__name__ == 'StaticText' and \
           type(parent).__name__ == 'TogglePanel':
            parent.Toggle(FakeEv(target))
            realYield()
            return

    w, h = target.GetClientSize().Get()
    x, y = target.GetScreenPosition()

    if pos is None:
        pos = [0.5, 0.5]

    x += w * pos[0]
    y += h * pos[1]

    sim.MouseMove(x, y)
    realYield()
    if   stype == 0: sim.MouseClick(btn)
    elif stype == 1: sim.MouseDblClick(btn)
    else:
        sim.MouseDown(btn)
        sim.MouseUp(btn)
    realYield()


def simtext(sim, target, text, enter=True):

    target.SetFocus()
    parent = target.GetParent()

    # The EVT_TEXT_ENTER event
    # does not seem to occur
    # under docker/GTK so we
    # have to hack. EVT_TEXT
    # does work though.
    if GTK and type(parent).__name__ == 'FloatSpinCtrl':
        if enter:
            target.ChangeValue(text)
            parent._FloatSpinCtrl__onText(None)
        else:
            target.SetValue(text)

    elif GTK and type(parent).__name__ == 'AutoTextCtrl':
        if enter:
            target.ChangeValue(text)
            parent._AutoTextCtrl__onEnter(None)
        else:
            target.SetValue(text)
    else:
        target.SetValue(text)

        if enter:
            sim.KeyDown(wx.WXK_RETURN)

    realYield()


def simkey(sim, target, key, down=True, up=False):

    class FakeEv(object):
        def __init__(self, obj, key):
            self.obj = obj
            self.key = key
        def GetKeyCode(self):
            return self.key
        def Skip(self):
            pass
        def GetEventObject(self):
            return self.obj
        def ResumePropagation(self, *a):
            pass


    parent = None
    if target is not None:
        target.SetFocus()
        parent = target.GetParent()

    if GTK:

        if down and type(parent).__name__ == 'AutoTextCtrl':
            parent._AutoTextCtrl__onKeyDown(FakeEv(target, key))

        elif down and type(parent).__name__ == 'FloatSpinCtrl':
            parent._FloatSpinCtrl__onKeyDown(FakeEv(target, key))
        elif down and type(parent).__name__ == 'TextTagPanel':
            if type(target).__name__ == 'AutoTextCtrl':
                parent._TextTagPanel__onNewTagKeyDown(FakeEv(target, key))
            elif type(target).__name__ == 'StaticTextTag':
                parent._TextTagPanel__onTagKeyDown(FakeEv(target, key))

        elif down and type(parent).__name__ == 'AutoCompletePopup':
            if type(target).__name__ == 'TextCtrl':
                parent._AutoCompletePopup__onKeyDown(FakeEv(target, key))
            elif type(target).__name__ == 'ListBox':
                parent._AutoCompletePopup__onListKeyDown(FakeEv(target, key))
        elif down:
            sim.KeyDown(key)

    elif down:
        sim.KeyDown(key)

    if up:
        sim.KeyUp(key)
    realYield()


def simfocus(from_, to):

    class FakeEv(object):
        def __init__(self):
            pass
        def Skip(self):
            pass

    if GTK:
        if type(from_).__name__ == 'FloatSpinCtrl':
            from_._FloatSpinCtrl__onKillFocus(FakeEv())

    to.SetFocus()
    realYield()



#!/usr/bin/env python
#
# numberdialog.py - An alternative to wx.NumberEntryDialog.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`NumberDialog` class, a dialog which
allows the user to enter a number.
"""

import wx


from . import floatspin


class NumberDialog(wx.Dialog):
    """A :class:`wx.Dialog` which prompts the user for a number.

    This class differs from the :class:`wx.NumberEntryDialog` in that
    it supports floating point numbers.

    A ``NumberDialog`` contains a :class:`wx.FloatSpinCtrl` and *Ok*/*Cancel*
    buttons, allowing the user to specify a number. If the user pushes the
    *Ok* button, the number they entered will be accessible via the
    :meth:`GetValue` method.
    """

    def __init__(self,
                 parent,
                 real=True,
                 title=None,
                 message=None,
                 initial=None,
                 minValue=None,
                 maxValue=None,
                 okText=None,
                 cancelText=None):
        """Create a :class:`NumberDialog`.

        :arg parent:     The :mod:`wx` parent object.

        :arg real:       If ``True``, a floating point number will
                         be accepted. Otherwise, only integers are
                         accepted.

        :arg title:      Dialog title.

        :arg message:    If not None, a :class:`wx.StaticText` label
                         is added, containing the message.

        :arg initial:    Initial value.

        :arg minValue:   Minimum value.

        :arg maxValue:   Maximum value.

        :arg okText:     Text for OK button. Defaults to "Ok".

        :arg cancelText: Text for Cancel button. Defaults to "Cancel"
        """

        if title      is None: title      = ''
        if initial    is None: initial    = 0
        if okText     is None: okText     = 'Ok'
        if cancelText is None: cancelText = 'Cancel'

        wx.Dialog.__init__(self, parent, title=title)

        self.__value     = None
        self.__cancelled = False
        self.__panel     = wx.Panel(self)
        self.__sizer     = wx.BoxSizer(wx.VERTICAL)
        self.__panel.SetSizer(self.__sizer)

        self.__buttonPanel = wx.Panel(self.__panel)
        self.__buttonSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__buttonPanel.SetSizer(self.__buttonSizer)

        if message is not None:
            self.__label = wx.StaticText(self.__panel, label=message)
        else:
            self.__label = (0, 0)

        style = floatspin.FSC_MOUSEWHEEL
        if not real:
            style |= floatspin.FSC_INTEGER
        if minValue is None and maxValue is None:
            style |= floatspin.FSC_NO_LIMIT

        self.__spinCtrl = floatspin.FloatSpinCtrl(
            self.__panel,
            minValue=minValue,
            maxValue=maxValue,
            value=initial,
            style=style)

        self.__okButton     = wx.Button(self.__buttonPanel, label=okText)
        self.__cancelButton = wx.Button(self.__buttonPanel, label=cancelText)

        self.__buttonSizer.Add(self.__okButton,
                               flag=wx.EXPAND | wx.ALL,
                               proportion=1,
                               border=2)
        self.__buttonSizer.Add(self.__cancelButton,
                               flag=wx.EXPAND | wx.ALL,
                               proportion=1,
                               border=2)

        self.__sizer.Add(self.__label, flag=wx.EXPAND | wx.ALL, border=10)
        self.__sizer.Add(self.__spinCtrl,
                         flag=(wx.EXPAND | wx.LEFT | wx.RIGHT),
                         border=15)

        self.__sizer.Add(self.__buttonPanel,
                         flag=wx.EXPAND | wx.ALL,
                         border=10)

        self.__spinCtrl    .Bind(wx.EVT_TEXT_ENTER, self.__onEnter)
        self.__okButton    .Bind(wx.EVT_BUTTON,     self.__onOk)
        self.__cancelButton.Bind(wx.EVT_BUTTON,     self.__onCancel)

        self.__panel.Layout()
        self.__panel.Fit()

        self.Fit()
        self.CentreOnParent()


    @property
    def floatSpinCtrl(self):
        """Returns a reference to the :class:`.FloatSpinCtrl`. """
        return self.__spinCtrl


    @property
    def okButton(self):
        """Returns a reference to the ok ``Button``. """
        return self.__okButton


    @property
    def cancelButton(self):
        """Returns a reference to the cancel ``Button``. """
        return self.__cancelButton


    def GetValue(self):
        """Return the value that the user entered.

        After a valid value has been entered, and OK button pushed (or
        enter pressed), this method may be used to retrieve the value.
        Returns ``None`` in all other situations.
        """
        if self.__cancelled: return None
        else:                return self.__value


    def __onEnter(self, ev):
        """Called when the enter key is pressed in the ``FloatSpinCtrl``.
        Forwards the event to the :meth:`__onOk` method.
        """
        # If the user typed in a value, and the
        # floatspin changed that value (e.g.
        # clamped it to min/max bounds), don't
        # close the dialog.
        if not ev.changed:
            self.__onOk(ev)


    def __onOk(self, ev):
        """Called when the Ok button is pushed. The entered value is stored
        and the dialog is closed.  The value may be retrieved via the
        :meth:`GetValue` method.
        """

        self.__value = self.__spinCtrl.GetValue()

        if self.IsModal():
            self.EndModal(wx.ID_OK)
        else:
            self.SetReturnCode(wx.ID_OK)
            self.Close()


    def __onCancel(self, ev):
        """Called when the Cancel button is pushed. Closes the dialog."""

        self.__value     = None
        self.__cancelled = True

        if self.IsModal():
            self.EndModal(wx.ID_CANCEL)
        else:
            self.SetReturnCode(wx.ID_CANCEL)
            self.Close()



#!/usr/bin/env python
#
# test_layout.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import numpy as np

import fsleyes_widgets.utils.layout as fsllayout


def test_padBitmap():

    bmp       = np.zeros((10, 10, 4))
    expectedh = np.ones( (10, 20, 4))
    expectedv = np.ones( (20, 10, 4))

    expectedh[ :,   5:15, :] = 0
    expectedv[5:15,  :,   :] = 0

    resulth = fsllayout.padBitmap(bmp, 20, 20, True,  (1, 1, 1, 1))
    resultv = fsllayout.padBitmap(bmp, 20, 20, False, (1, 1, 1, 1))

    assert np.all(resulth == expectedh)
    assert np.all(resultv == expectedv)


def test_buildCanvasBox_withLabels():

    labelbmps = {
        'left'   : np.zeros((5, 5, 4)),
        'right'  : np.zeros((5, 5, 4)),
        'top'    : np.zeros((5, 5, 4)),
        'bottom' : np.zeros((5, 5, 4))
    }

    canvasbmp               = np.ones((20, 20, 4))
    expected                = np.zeros((30, 30, 4))
    expected[5:25, 5:25, :] = canvasbmp

    box = fsllayout.buildCanvasBox(canvasbmp, labelbmps, True, 0)
    bmp = fsllayout.layoutToBitmap(box)

    assert np.all(bmp == expected)


def test_buildCanvasBox_noLabels():

    canvasbmp = np.ones((5, 5, 4))

    box = fsllayout.buildCanvasBox(canvasbmp, None, False, 0)

    bmp = fsllayout.layoutToBitmap(box)

    assert np.all(bmp == canvasbmp)


def test_buildOrthoLayout_vertical():

    canvasbmps = [
        np.zeros((10, 10, 4)),
        np.zeros((10, 10, 4)),
        np.zeros((10, 10, 4))]

    canvasbmps[0][:] = 1
    canvasbmps[1][:] = 2
    canvasbmps[2][:] = 3

    expected = np.vstack(canvasbmps)

    result = fsllayout.buildOrthoLayout(canvasbmps, None, 'vertical', False, 0)
    result = fsllayout.layoutToBitmap(result)

    assert np.all(result == expected)


def test_buildOrthoLayout_horizontal():

    canvasbmps = [
        np.zeros((10, 10, 4)),
        np.zeros((10, 10, 4)),
        np.zeros((10, 10, 4))]

    canvasbmps[0][:] = 1
    canvasbmps[1][:] = 2
    canvasbmps[2][:] = 3

    expected = np.hstack(canvasbmps)

    result = fsllayout.buildOrthoLayout(
        canvasbmps, None, 'horizontal', False, 0)
    result = fsllayout.layoutToBitmap(result)

    assert np.all(result == expected)


def test_buildOrthoLayout_grid():

    canvasbmps = [
        np.zeros((10, 10, 4)),
        np.zeros((10, 10, 4)),
        np.zeros((10, 10, 4))]

    canvasbmps[0][:] = 1
    canvasbmps[1][:] = 2
    canvasbmps[2][:] = 3

    expected = np.zeros((20, 20, 4))
    expected[:10,  :10,  :] = canvasbmps[0]
    expected[:10,   10:, :] = canvasbmps[1]
    expected[ 10:, :10,  :] = canvasbmps[2]

    result = fsllayout.buildOrthoLayout(
        canvasbmps, None, 'grid', False, 0)
    result = fsllayout.layoutToBitmap(result)

    assert np.all(result == expected)


def test_calcSizes():

    # calcSizes  is just a wrapper around the
    # calcGrid/Horizontal/VerticalSizes functions

    # For grid layout, canvas edges have
    # to line up - see calcGridSizes docs
    gaxes = [(0, 2), (1, 2), (0, 1)]
    axes  = [(0, 1), (0, 2), (1, 2)]

    # bounds, width, height
    testcases = [
        ([1.0, 1.0, 1.0], 200, 200),
        ([1.0, 1.0, 1.0], 100, 200),
        ([1.0, 1.0, 1.0], 200, 100),
        ([1.0, 2.0, 3.0], 200, 200)]

    for testcase in testcases:

        bounds, width, height = testcase

        hexp = fsllayout.calcHorizontalSizes(    axes,  bounds, width, height)
        vexp = fsllayout.calcVerticalSizes(      axes,  bounds, width, height)
        gexp = fsllayout.calcGridSizes(          gaxes, bounds, width, height)
        hres = fsllayout.calcSizes('horizontal', axes,  bounds, width, height)
        vres = fsllayout.calcSizes('vertical',   axes,  bounds, width, height)
        gres = fsllayout.calcSizes('grid',       gaxes, bounds, width, height)

        assert np.all(np.isclose(hexp, hres))
        assert np.all(np.isclose(vexp, vres))
        assert np.all(np.isclose(gexp, gres))


def test_calcGridSizes():

    axes = [(0, 2), (1, 2), (0, 1)]

    # bounds, width, height, expected
    testcases = [
        ([1.0, 1.0, 1.0], 200, 200, [(100, 100),
                                     (100, 100),
                                     (100, 100)]),

        ([1.0, 1.0, 1.0], 100, 200, [(50, 100),
                                     (50, 100),
                                     (50, 100)]),

        ([1.0, 1.0, 1.0], 200, 100, [(100, 50),
                                     (100, 50),
                                     (100, 50)]),

        ([1.0, 2.0, 3.0], 200, 200, [(200 * 1 / 3.0, 200 * 3 / 5.0),
                                     (200 * 2 / 3.0, 200 * 3 / 5.0),
                                     (200 * 1 / 3.0, 200 * 2 / 5.0)]),
    ]

    for testcase in testcases:
        bounds, width, height = testcase[:3]
        expected              = testcase[ 3]

        result = fsllayout.calcGridSizes(axes, bounds, width, height)

        assert np.all(np.isclose(result, expected))

        # Passing less than three canvases should be
        # equivalent to calcHorizontalSizes
        for ncanvases in [1, 2]:
            naxes = axes[:ncanvases]

            gres = fsllayout.calcGridSizes(      naxes, bounds, width, height)
            hres = fsllayout.calcHorizontalSizes(naxes, bounds, width, height)

            assert np.all(np.isclose(gres, hres))


def test_calcHorizontalSizes():
    axes = [(0, 1), (0, 2), (1, 2)]

    # bonds, width, height, expected
    testcases = [
        ([1.0, 1.0, 1.0], 300, 100, [(100, 100),
                                     (100, 100),
                                     (100, 100)]),

        ([1.0, 1.0, 1.0], 100, 100, [(100 / 3.0, 100),
                                     (100 / 3.0, 100),
                                     (100 / 3.0, 100)]),

        ([1.0, 1.0, 1.0], 100, 200, [(100 / 3.0, 200),
                                     (100 / 3.0, 200),
                                     (100 / 3.0, 200)]),

        ([1.0, 2.0, 3.0], 100, 100, [(100 * 1 / 4.0, 100 * 2 / 3.0),
                                     (100 * 1 / 4.0, 100),
                                     (100 * 2 / 4.0, 100)])
    ]

    for testcase in testcases:
        bounds, width, height = testcase[:3]
        expected              = testcase[ 3]

        result = fsllayout.calcHorizontalSizes(axes, bounds, width, height)

        assert np.all(np.isclose(result, expected))


def test_calcVerticalSizes():

    axes = [(0, 1), (0, 2), (1, 2)]

    # bonds, width, height, expected
    testcases = [
        ([1.0, 1.0, 1.0], 100, 300, [(100, 100),
                                     (100, 100),
                                     (100, 100)]),

        ([1.0, 1.0, 1.0], 100, 100, [(100, 100 / 3.0),
                                     (100, 100 / 3.0),
                                     (100, 100 / 3.0)]),


        ([1.0, 1.0, 1.0], 200, 100, [(200, 100 / 3.0),
                                     (200, 100 / 3.0),
                                     (200, 100 / 3.0)]),

        ([1.0, 2.0, 3.0], 100, 100, [(50,  100 * 2.0 / 8),
                                     (50,  100 * 3.0 / 8),
                                     (100, 100 * 3.0 / 8)]),
    ]

    for testcase in testcases:
        bounds, width, height = testcase[:3]
        expected              = testcase[ 3]

        result = fsllayout.calcVerticalSizes(axes, bounds, width, height)

        assert np.all(np.isclose(result, expected))



def test_calcPixWidth():

    # (wldWidth, wldHeight, pixHeight, expected)
    testcases = [
        (1.0, 1.0, 100, 100),
        (1.0, 2.0, 100,  50),
        (2.0, 1.0,  50, 100)]

    for wldWidth, wldHeight, pixHeight, exp in testcases:
        assert fsllayout.calcPixWidth(wldWidth, wldHeight, pixHeight) == exp


def test_calcPixHeight():

    # (wldWidth, wldHeight, pixWidth, expected)
    testcases = [
        (1.0, 1.0, 100, 100),
        (1.0, 2.0,  50, 100),
        (2.0, 1.0, 100,  50)]

    for wldWidth, wldHeight, pixHeight, exp in testcases:
        assert fsllayout.calcPixHeight(wldWidth, wldHeight, pixHeight) == exp



#!/usr/bin/env python
#
# test_bitmapradio.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path as op

import pytest

import wx

from . import run_with_wx, simclick

import fsleyes_widgets.bitmapradio  as bmpradio
import fsleyes_widgets.bitmaptoggle as bmptoggle


datadir = op.join(op.dirname(__file__), 'testdata', 'bitmapbuttons')


def test_Create():
    run_with_wx(_test_Create)
def _test_Create():

    falseicon = op.join(datadir, 'false.png')
    trueicon  = op.join(datadir, 'true.png')
    falseicon = wx.Bitmap(falseicon, wx.BITMAP_TYPE_PNG)
    trueicon  = wx.Bitmap(trueicon,  wx.BITMAP_TYPE_PNG)

    frame = wx.GetApp().GetTopWindow()

    btn1 = bmpradio.BitmapRadioBox(frame)
    btn2 = bmpradio.BitmapRadioBox(frame, style=wx.HORIZONTAL)
    btn3 = bmpradio.BitmapRadioBox(frame, style=wx.VERTICAL)

    for btn in [btn1, btn2, btn3]:

        with pytest.raises(IndexError):
            btn.SetSelection(0)

        btn.Set([trueicon, falseicon])
        btn.SetSelection(0)
        assert btn.GetSelection() == 0
        btn.SetSelection(1)
        assert btn.GetSelection() == 1

        btn.EnableChoice(0)
        btn.EnableChoice(0, False)
        btn.DisableChoice(0)

        with pytest.raises(IndexError):
            btn.EnableChoice(2)

        with pytest.raises(IndexError):
            btn.SetSelection(2)

        btn.Clear()
        with pytest.raises(IndexError):
            btn.SetSelection(0)

        btn.AddChoice(trueicon,  falseicon)
        btn.AddChoice(falseicon, trueicon)

        btn.SetSelection(0)
        assert btn.GetSelection() == 0
        btn.SetSelection(1)
        assert btn.GetSelection() == 1

        with pytest.raises(IndexError):
            btn.SetSelection(2)


def test_Event():
    run_with_wx(_test_Event)
def _test_Event():
    falseicon = wx.Bitmap(op.join(datadir, 'false.png'), wx.BITMAP_TYPE_PNG)
    trueicon  = wx.Bitmap(op.join(datadir, 'true.png'),  wx.BITMAP_TYPE_PNG)

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = wx.Panel(frame)
    rad   = bmpradio.BitmapRadioBox(panel)
    szr   = wx.BoxSizer(wx.HORIZONTAL)
    szr.Add(rad, flag=wx.EXPAND)
    panel.SetSizer(szr)

    result = [None]

    def handler(ev):
        result[0] = (ev.index, ev.clientData)

    rad.Bind(bmpradio.EVT_BITMAP_RADIO_EVENT, handler)

    rad.Set([trueicon, falseicon], ['true', 'false'])

    panel.Layout()
    panel.Fit()
    wx.Yield()

    btns = []
    for c in rad.GetChildren():
        if isinstance(c, bmptoggle.BitmapToggleButton):
            btns.append(c)

    simclick(sim, btns[0])
    assert result[0] == (0, 'true')
    simclick(sim, btns[1])
    assert result[0] == (1, 'false')



#!/usr/bin/env python
#
# elistbox.py - An alternative to wx.gizmos.EditableListBox.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

"""This module provides the :class:`EditableListBox` class, an alternative to
:class:`wx.gizmos.EditableListBox`.
"""

import math
import logging

import six

import wx
import wx.lib.newevent as wxevent
import wx.lib.stattext as stattext


log = logging.getLogger(__name__)


class EditableListBox(wx.Panel):
    """A panel which displays a list of items.


    An ``EditableListBox`` contains a :class:`wx.Panel` which in turn contains
    a collection of :class:`wx.StaticText` widgets, which are laid out
    vertically, and display labels for each of the items in the list. Some
    rudimentary wrapper methods for modifying the list contents are provided
    by an ``EditableListBox`` object, with an interface similar to that of the
    :class:`wx.ListBox` class.


    In addition to displaying ``StaticText`` controls, the ``EditableListBox``
    can also display arbitrary panels/controls associated with each label -
    see the :meth:`Insert` and :meth:`SetItemWidget`` methods.


     .. warning:: If you are using an ``EditableListBox`` to display arbitrary
                  controls/panels it is important to know that the
                  ``EditableListBox`` assumes that all items are of the same
                  size. Sizing/scrolling will not work properly if
                  controls/panels for different list items are of different
                  sizes.


    The following style flags are available:

     .. autosummary::
        ELB_NO_SCROLL
        ELB_NO_ADD
        ELB_NO_REMOVE
        ELB_NO_MOVE
        ELB_REVERSE
        ELB_TOOLTIP
        ELB_EDITABLE
        ELB_NO_LABELS
        ELB_WIDGET_RIGHT
        ELB_TOOLTIP_DOWN
        ELB_SCROLL_BUTTONS


    An ``EditableListBox`` generates the following events:

     .. autosummary::
        ListSelectEvent
        ListAddEvent
        ListRemoveEvent
        ListMoveEvent
        ListEditEvent
        ListDblClickEvent


     .. note::
        The ``EditableListBox`` is an alternative to the
        :class:`wx.gizmos.EditableListBox`. The latter is buggy under OS X,
        and getting tooltips working with the :class:`wx.ListBox` is an
        absolute pain in the behind. So I felt the need to replicate its
        functionality.  This implementation supports single selection only.
    """

    _selectedFG = '#000000'
    """Default foreground colour for the currently selected item."""


    _defaultFG = '#000000'
    """Default foreground colour for unselected items."""


    _selectedBG = '#cdcdff'
    """Background colour for the currently selected item."""


    _defaultBG  = '#FFFFFF'
    """Background colour for the unselected items."""


    def __init__(
            self,
            parent,
            labels=None,
            clientData=None,
            tooltips=None,
            style=0):
        """Create an ``EditableListBox``.

        :arg parent:     :mod:`wx` parent object

        :arg labels:     List of strings, the items in the list

        :arg clientData: List of data associated with the list items.

        :arg tooltips:   List of strings, tooltips for each item.

        :arg style:      Style bitmask - accepts :data:`ELB_NO_SCROLL`,
                         :data:`ELB_NO_ADD`, :data:`ELB_NO_REMOVE`,
                         :data:`ELB_NO_MOVE`, :data:`ELB_REVERSE`,
                         :data:`ELB_TOOLTIP`, :data:`ELB_EDITABLE`,
                         :data:`ELB_NO_LABEL`, :data:`ELB_WIDGET_RIGHT`,
                         :data:`ELB_TOOLTIP_DOWN`, and
                         :data:`ELB_SCROLL_BUTTONS`.
        """

        wx.Panel.__init__(self, parent, style=wx.WANTS_CHARS)

        reverseOrder  =      style & ELB_REVERSE
        addScrollbar  = not (style & ELB_NO_SCROLL)
        addSupport    = not (style & ELB_NO_ADD)
        removeSupport = not (style & ELB_NO_REMOVE)
        moveSupport   = not (style & ELB_NO_MOVE)
        editSupport   =      style & ELB_EDITABLE
        showTooltips  =      style & ELB_TOOLTIP
        noLabels      =      style & ELB_NO_LABELS
        widgetOnRight =      style & ELB_WIDGET_RIGHT
        tooltipDown   =      style & ELB_TOOLTIP_DOWN   and not showTooltips
        scrollButtons =      style & ELB_SCROLL_BUTTONS and     addScrollbar
        noButtons     = not any((addSupport, removeSupport, moveSupport))

        if noLabels:
            editSupport = False

        self.__reverseOrder  = reverseOrder
        self.__showTooltips  = showTooltips
        self.__moveSupport   = moveSupport
        self.__editSupport   = editSupport
        self.__noLabels      = noLabels
        self.__widgetOnRight = widgetOnRight
        self.__tooltipDown   = tooltipDown
        self.__scrollButtons = scrollButtons

        if labels     is None: labels     = []
        if clientData is None: clientData = [None] * len(labels)
        if tooltips   is None: tooltips   = [None] * len(labels)

        # index of the currently selected
        # item, and the list of items
        # (_ListItem instances).
        self.__selection  = wx.NOT_FOUND
        self.__listItems  = []

        # the panel containing the list items
        # This is laid out with two sizers -
        # the sizer contains the list items, and
        # the SizerSizer contains scroll buttons
        # (if enabled) and the item sizer.
        self.__listPanel      = wx.Panel(self, style=wx.WANTS_CHARS)
        self.__listSizerSizer = wx.BoxSizer(wx.VERTICAL)
        self.__listSizer      = wx.BoxSizer(wx.VERTICAL)

        self.__listSizerSizer.Add(self.__listSizer,
                                  flag=wx.EXPAND,
                                  proportion=1)

        self.__listPanel.SetSizer(self.__listSizerSizer)
        self.__listPanel.SetBackgroundColour(EditableListBox._defaultBG)

        if addScrollbar:
            self.__scrollbar = wx.ScrollBar(self, style=wx.SB_VERTICAL)
        else:
            self.__scrollbar = None

        # A panel containing buttons for doing stuff with the list
        if not noButtons:
            self.__buttonPanel      = wx.Panel(self)
            self.__buttonPanelSizer = wx.BoxSizer(wx.VERTICAL)
            self.__buttonPanel.SetSizer(self.__buttonPanelSizer)

        # Buttons for moving the selected item up/down
        if moveSupport:
            self.__upButton   = wx.Button(self.__buttonPanel,
                                          label=six.u('\u25B2'),
                                          style=wx.BU_EXACTFIT)
            self.__downButton = wx.Button(self.__buttonPanel,
                                          label=six.u('\u25BC'),
                                          style=wx.BU_EXACTFIT)
            self.__upButton  .Bind(wx.EVT_BUTTON, self.__moveItemUp)
            self.__downButton.Bind(wx.EVT_BUTTON, self.__moveItemDown)

            self.__buttonPanelSizer.Add(self.__upButton,   flag=wx.EXPAND)
            self.__buttonPanelSizer.Add(self.__downButton, flag=wx.EXPAND)

        # Button for adding new items
        if addSupport:
            self.__addButton = wx.Button(self.__buttonPanel, label='+',
                                         style=wx.BU_EXACTFIT)
            self.__addButton.Bind(wx.EVT_BUTTON, self.__addItem)
            self.__buttonPanelSizer.Add(self.__addButton, flag=wx.EXPAND)

        # Button for removing the selected item
        if removeSupport:
            self.__removeButton = wx.Button(self.__buttonPanel, label='-',
                                            style=wx.BU_EXACTFIT)

            self.__removeButton.Bind(wx.EVT_BUTTON, self.__removeItem)
            self.__buttonPanelSizer.Add(self.__removeButton, flag=wx.EXPAND)

        # Up/down scroll buttons above/below the list
        if self.__scrollButtons:

            # Using wx.lib.stattext instead
            # of wx.StaticText because GTK
            # can't horizontally align text.
            self.__scrollUp   = stattext.GenStaticText(self.__listPanel,
                                                       label=six.u('\u25B2'),
                                                       style=wx.ALIGN_CENTRE)
            self.__scrollDown = stattext.GenStaticText(self.__listPanel,
                                                       label=six.u('\u25BC'),
                                                       style=wx.ALIGN_CENTRE)

            self.__scrollUp  .SetFont(self.__scrollUp  .GetFont().Smaller())
            self.__scrollDown.SetFont(self.__scrollDown.GetFont().Smaller())

            self.__listSizerSizer.Insert(0, self.__scrollUp,   flag=wx.EXPAND)
            self.__listSizerSizer.Insert(2, self.__scrollDown, flag=wx.EXPAND)

            self.__scrollUp  .Enable(False)
            self.__scrollDown.Enable(False)

            self.__scrollUp  .SetBackgroundColour('#e0e0e0')
            self.__scrollDown.SetBackgroundColour('#e0e0e0')

            self.__scrollUp  .Bind(wx.EVT_LEFT_UP, self.__onScrollButton)
            self.__scrollDown.Bind(wx.EVT_LEFT_UP, self.__onScrollButton)

        else:
            self.__scrollUp   = None
            self.__scrollDown = None

        self.__sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.SetSizer(self.__sizer)

        if not noButtons:
            self.__sizer.Add(self.__buttonPanel, flag=wx.EXPAND)

        self.__sizer.Add(self.__listPanel, flag=wx.EXPAND, proportion=1)

        if addScrollbar:
            self.__sizer.Add(self.__scrollbar, flag=wx.EXPAND)
            self.__scrollbar.Bind(wx.EVT_SCROLL,     self.__drawList)
            self            .Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)
            self.__listPanel.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)

        def refresh(ev):
            self.__updateScrollbar()
            self.__drawList()
            ev.Skip()

        # We use CHAR_HOOK for key events,
        # because we want to capture key
        # presses whenever this panel or
        # any of its children has focus.
        self.Bind(wx.EVT_CHAR_HOOK, self.__onKeyboard)
        self.Bind(wx.EVT_PAINT,     refresh)
        self.Bind(wx.EVT_SIZE,      refresh)

        for label, data, tooltip in zip(labels, clientData, tooltips):
            self.Append(label, data, tooltip)

        # Figure out a nice minimum width
        dummyLabel = wx.StaticText(self, label='w' * 25)
        lblWidth, lblHeight = dummyLabel.GetBestSize().Get()
        dummyLabel.Destroy()

        # If there are buttons on this listbox,
        # set the minimum height to the height
        # of said buttons, or the height of
        # four items, whichever is bigger
        if not noButtons:
            self.SetMinSize((
                lblWidth,
                max(4 * lblHeight, self.__buttonPanelSizer.CalcMin()[1])))

        # Otherwise set the minimum height
        # to the height of four items
        else:
            self.SetMinSize((lblWidth, 4 * lblHeight))

        self.Layout()


    def __onKeyboard(self, ev):
        """Called when a key is pressed. On up/down arrow key presses,
        changes the selected item, and scrolls if necessary.
        """

        key = ev.GetKeyCode()

        # GTK seemingly randomly steals focus from
        # this control, and gives it to something
        # else, unless we force-retain focus.
        wx.CallAfter(self.SetFocus)

        # We're only interested in
        # up/down key presses
        if ev.HasModifiers() or (key not in (wx.WXK_UP, wx.WXK_DOWN)):
            ev.Skip()
            return

        # On up/down keys, we want to
        # select the next/previous item
        if   key == wx.WXK_UP:   offset = -1
        elif key == wx.WXK_DOWN: offset =  1

        selected = self.__selection + offset

        if any((selected < 0, selected >= self.GetCount())):
            return

        # Change the selected item, simulating
        # a mouse click so that event listeners
        # are notified
        self.__itemClicked(None, self.__listItems[selected].labelWidget)

        # Update the scrollbar position, to make
        # sure the newly selected item is visible
        if self.__scrollbar is not None:
            scrollPos = self.__scrollbar.GetThumbPosition()

            if any((selected <  scrollPos,
                    selected >= scrollPos + self.__scrollbar.GetPageSize())):
                self.__onMouseWheel(None, -offset)


    def __onMouseWheel(self, ev=None, move=None):
        """Called when the mouse wheel is scrolled over the list. Scrolls
        through the list accordingly.

        :arg ev:   A :class:`wx.MouseEvent`

        :arg move: If called programmatically, a number indicating the
                   direction in which to scroll.
        """

        if self.__scrollbar is None:
            return

        if ev is not None:
            move = ev.GetWheelRotation()

        scrollPos = self.__scrollbar.GetThumbPosition()

        if   move < 0: self.__scrollbar.SetThumbPosition(scrollPos + 1)
        elif move > 0: self.__scrollbar.SetThumbPosition(scrollPos - 1)

        self.__drawList()
        self.SetFocus()


    def __onScrollButton(self, ev):
        """Called when either of the scroll up/down buttons are clicked (if
        the :data:`.ELB_SCROLL_BUTTONS` style is active). Scrolls the list
        up/down, if possible.
        """

        button = ev.GetEventObject()

        if not button.IsEnabled():
            return

        if button is self.__scrollUp: move =  1
        else:                         move = -1

        self.__onMouseWheel(move=move)


    def VisibleItemCount(self):
        """Returns the number of items in the list which are visible
        (i.e. which have not been hidden via a call to :meth:`ApplyFilter`).
        """
        nitems = 0

        for item in self.__listItems:
            if not item.hidden:
                nitems += 1

        return nitems


    def __drawList(self, ev=None):
        """'Draws' the set of items in the list according to the
        current scrollbar thumb position.
        """

        nitems = self.VisibleItemCount()

        if self.__scrollbar is not None:
            thumbPos     = self.__scrollbar.GetThumbPosition()
            itemsPerPage = self.__scrollbar.GetPageSize()
        else:
            thumbPos      = 0
            itemsPerPage = len(self.__listItems)

        if itemsPerPage >= nitems:
            start = 0
            end   = nitems
        else:
            start = thumbPos
            end   = thumbPos + itemsPerPage

        if end > nitems:

            start = start - (end - nitems)
            end   = nitems

        visI = 0
        for i, item in enumerate(self.__listItems):

            if item.hidden:
                self.__listSizer.Show(item.container, False)
                continue

            if (visI < start) or (visI >= end):
                self.__listSizer.Show(item.container, False)
            else:
                self.__listSizer.Show(item.container, True)

            visI += 1

        if self.__scrollButtons:
            self.__scrollUp  .Enable(thumbPos > 0)
            self.__scrollDown.Enable(end      < nitems)

        self.__listSizer.Layout()

        if ev is not None:
            ev.Skip()


    def __updateScrollbar(self, ev=None):
        """Updates the scrollbar parameters according to the number of items
        in the list, and the screen size of the list panel. If there is
        enough room to display all items in the list, the scroll bar is
        hidden.
        """

        if self.__scrollbar is None:
            return

        nitems     = self.VisibleItemCount()
        pageHeight = self.GetSize().GetHeight() - 5
        if self.__scrollUp is not None:
            pageHeight -= self.__scrollUp.GetSize().GetHeight()
        if self.__scrollDown is not None:
            pageHeight -= self.__scrollDown.GetSize().GetHeight()

        # Yep, I'm assuming that all
        # items are the same size
        if nitems > 0:
            sizer      = self.__listItems[0].container.GetSizer()
            itemHeight = sizer.CalcMin().GetHeight()
        else:
            itemHeight = 0

        if pageHeight == 0 or itemHeight == 0:
            itemsPerPage = nitems
        else:
            itemsPerPage = math.floor(pageHeight / float(itemHeight))

        thumbPos     = self.__scrollbar.GetThumbPosition()
        itemsPerPage = min(itemsPerPage, nitems)

        # Hide the scrollbar if there is enough
        # room to display the entire list (but
        # configure the scrollbar correctly)
        if nitems == 0 or itemsPerPage >= nitems:
            self.__scrollbar.SetScrollbar(0,
                                          nitems,
                                          nitems,
                                          nitems,
                                          True)
            self.__sizer.Show(self.__scrollbar, False)
        else:
            self.__sizer.Show(self.__scrollbar, True)
            self.__scrollbar.SetScrollbar(thumbPos,
                                          itemsPerPage,
                                          nitems,
                                          itemsPerPage,
                                          True)
        self.Layout()


    def __fixIndex(self, idx):
        """If the :data:`ELB_REVERSE` style is active, this method will return
        an inverted version of the given index. Otherwise it returns the index
        value unchanged.
        """

        if idx is None:               return idx
        if idx == wx.NOT_FOUND:       return idx
        if not self.__reverseOrder:   return idx

        fixIdx = len(self.__listItems) - idx - 1

        # if len(self.__listItems) is passed to Insert
        # (i.e. an item is to be appended to the list)
        # the above formula will produce -1
        if (idx == len(self.__listItems)) and (fixIdx == -1):
            fixIdx = 0

        return fixIdx


    def GetCount(self):
        """Returns the number of items in the list."""
        return len(self.__listItems)


    def Clear(self):
        """Removes all items from the list."""

        nitems = len(self.__listItems)
        for i in range(nitems):
            self.Delete(0)


    def ClearSelection(self):
        """Ensures that no items are selected."""

        for i, item in enumerate(self.__listItems):
            item.labelWidget.SetForegroundColour(item.defaultFGColour)
            item.labelWidget.SetBackgroundColour(item.defaultBGColour)
            item.container  .SetBackgroundColour(item.defaultBGColour)

            item.labelWidget.Refresh()
            item.container  .Refresh()

            if item.extraWidget is not None:
                item.extraWidget.SetBackgroundColour(item.defaultBGColour)
                item.extraWidget.Refresh()

        self.__selection = wx.NOT_FOUND


    def SetSelection(self, n):
        """Selects the item at the given index."""

        if n != wx.NOT_FOUND and (n < 0 or n >= len(self.__listItems)):
            raise IndexError('Index {} out of bounds'.format(n))

        self.ClearSelection()

        if n == wx.NOT_FOUND: return

        self.__selection = self.__fixIndex(n)

        item = self.__listItems[self.__selection]

        item.labelWidget.SetForegroundColour(item.selectedFGColour)
        item.labelWidget.SetBackgroundColour(item.selectedBGColour)
        item.container  .SetBackgroundColour(item.selectedBGColour)

        item.labelWidget.Refresh()
        item.container  .Refresh()

        if item.extraWidget is not None:
            item.extraWidget.SetBackgroundColour(item.selectedBGColour)
            item.extraWidget.Refresh()

        self.__updateMoveButtons()


    def GetSelection(self):
        """Returns the index of the selected item, or :data:`wx.NOT_FOUND`
        if no item is selected.
        """
        return self.__fixIndex(self.__selection)


    def Insert(self,
               label,
               pos,
               clientData=None,
               tooltip=None,
               extraWidget=None):
        """Insert an item into the list.

        :arg label:       The label to be displayed.

        :arg pos:         Index at which the item is to be inserted.

        :arg clientData:  Data associated with the item.

        :arg tooltip:     Tooltip to be shown, if the :data:`ELB_TOOLTIP`
                          style is active.

        :arg extraWidget: A widget to be displayed alongside the label.
        """

        if pos < 0 or pos > self.GetCount():
            raise IndexError('Index {} out of bounds'.format(pos))

        pos = self.__fixIndex(pos)

        if self.__noLabels:
            label = ''

        # StaticText under Linux/GTK poses problems -
        # we cannot set background colour, nor can we
        # intercept mouse motion events. So we embed
        # the StaticText widget within a wx.Panel.
        container   = wx.Panel(self.__listPanel, style=wx.WANTS_CHARS)
        labelWidget = wx.StaticText(container,
                                    label=label,
                                    style=(wx.ST_ELLIPSIZE_MIDDLE   |
                                           wx.ALIGN_CENTRE_VERTICAL |
                                           wx.WANTS_CHARS))

        sizer = wx.BoxSizer(wx.HORIZONTAL)
        container.SetSizer(sizer)

        sizerItems = [labelWidget]
        if self.__noLabels: sizerFlags = [{}]
        else:               sizerFlags = [{'flag'       : wx.ALIGN_CENTRE,
                                           'proportion' : 1}]

        if extraWidget is not None:
            extraWidget.Reparent(container)

            if self.__widgetOnRight: index = 1
            else:                    index = 0

            sizerItems.insert(index, extraWidget)
            sizerFlags.insert(index, {})

        for item, flags in zip(sizerItems, sizerFlags):
            sizer.Add(item, **flags)

        labelWidget.Bind(wx.EVT_LEFT_DOWN, self.__itemClicked)
        container  .Bind(wx.EVT_LEFT_DOWN, self.__itemClicked)

        # Under linux/GTK, mouse wheel handlers
        # need to be added to children, not
        # just the top level container
        if self.__scrollbar is not None:
            labelWidget.Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)
            container  .Bind(wx.EVT_MOUSEWHEEL, self.__onMouseWheel)

        item = _ListItem(label,
                         clientData,
                         tooltip,
                         labelWidget,
                         container,
                         EditableListBox._defaultFG,
                         EditableListBox._selectedFG,
                         EditableListBox._defaultBG,
                         EditableListBox._selectedBG,
                         extraWidget)

        def onEdit(ev):
            self.__onEdit(ev, item)

        def onDblClick(ev):
            self.__onDoubleClick(ev, item)

        # If the items are editable,
        # double clicking will call
        # the __onEdit method
        if self.__editSupport:
            labelWidget.Bind(wx.EVT_LEFT_DCLICK, onEdit)
            container  .Bind(wx.EVT_LEFT_DCLICK, onEdit)

        # Otherwise, double clicking will
        # call the __onDoubleClick method
        else:
            labelWidget.Bind(wx.EVT_LEFT_DCLICK, onDblClick)
            container  .Bind(wx.EVT_LEFT_DCLICK, onDblClick)

        log.debug('Inserting item ({}) at index {}'.format(label, pos))

        self.__listItems.insert(pos, item)
        self.__listSizer.Insert(pos, container, flag=wx.EXPAND)
        self.__listSizer.Layout()

        # if an item was inserted before the currently
        # selected item, the __selection index will no
        # longer be valid - fix it.
        if self.__selection != wx.NOT_FOUND and pos < self.__selection:
            self.__selection = self.__selection + 1

        # Make sure item fg/bg colours are up to date
        self.SetSelection(self.__fixIndex(self.__selection))

        self.__updateMoveButtons()
        if self.__tooltipDown: self.__configTooltipDown(item)
        else:                  self.__configTooltip(    item)

        # Make sure the enabled state of the
        # new label/widget is consistent with
        # the state of this elistbox.
        container.Enable(self.IsEnabled())

        self.Refresh()


    def __configTooltip(self, listItem):
        """If the :data:`ELB_TOOLTIP` style was enabled, this method
        configures mouse-over listeners on the widget representing the given
        list item, so the item displays the tool tip on mouse overs.

        If :data:`ELB_TOOLTIP` is not enabled, a regular tooltip is configured.
        """

        if not self.__showTooltips:
            if listItem.tooltip is not None:
                tooltip = wx.ToolTip(listItem.tooltip)
                listItem.labelWidget.SetToolTip(tooltip)

        else:

            def mouseOver(ev):
                if listItem.tooltip is not None:
                    listItem.labelWidget.SetLabel(listItem.tooltip)

            def mouseOut(ev):
                listItem.labelWidget.SetLabel(listItem.label)

            # Register motion listeners on the widget
            # container so it works under GTK
            listItem.container.Bind(wx.EVT_ENTER_WINDOW, mouseOver)
            listItem.container.Bind(wx.EVT_LEAVE_WINDOW, mouseOut)


    def __configTooltipDown(self, listItem):
        """If the :data:`ELB_TOOLTIP_DOWN` style was enabled, this method
        configures mouse-down listeners on the given
        list item widget, so the item displays the tool tip on mouse down.

        This method is not called if :data:`ELB_TOOLTIP_DOWN` is not enabled.
        """

        # The tooltip is shown only after the mouse
        # has been held down for a short period of
        # time. This is required so the tooltip is
        # not shown on regular clicks/double clicks.
        listItem._cancelTooltipDown = False

        def changeLabel(lbl):
            if not listItem._cancelTooltipDown:
                listItem.labelWidget.SetLabel(lbl)

        def mouseDown(ev):
            ev.Skip()
            listItem._cancelTooltipDown = False
            if listItem.tooltip is not None:
                wx.CallLater(300, changeLabel, listItem.tooltip)

        def mouseUp(ev):

            ev.Skip()
            listItem._cancelTooltipDown = True

            listItem.labelWidget.SetLabel(listItem.label)

        listItem.labelWidget.Bind(wx.EVT_LEFT_DOWN, mouseDown)
        listItem.labelWidget.Bind(wx.EVT_LEFT_UP,   mouseUp)


    def Append(self, label, clientData=None, tooltip=None, extraWidget=None):
        """Appends an item to the end of the list.

        :arg label:       The label to be displayed

        :arg clientData:  Data associated with the item

        :arg tooltip:     Tooltip to be shown, if the :data:`ELB_TOOLTIP`
                          style is active.

        :arg extraWidget: A widget to be displayed alonside the item.
        """
        self.Insert(label,
                    len(self.__listItems),
                    clientData,
                    tooltip,
                    extraWidget)


    def Delete(self, n):
        """Removes the item at the given index from the list."""

        n = self.__fixIndex(n)

        if n < 0 or n >= len(self.__listItems):
            raise IndexError('Index {} out of bounds'.format(n))

        item = self.__listItems.pop(n)

        self.__listSizer.Remove(n)

        # Destroying the container will result in the
        # child widget(s) being destroyed as well.
        item.container.Destroy()

        self.__listSizer.Layout()

        # if the deleted item was selected, clear the selection
        if self.__selection == n:
            self.ClearSelection()

        # or if the deleted item was before the
        # selection, fix the selection index
        elif self.__selection > n:
            self.__selection = self.__selection - 1

        self.__updateMoveButtons()
        self.__updateScrollbar()
        self.Refresh()


    def IndexOf(self, clientData):
        """Returns the index of the list item with the specified
        ``clientData``.
        """

        for i, item in enumerate(self.__listItems):
            if item.data == clientData:
                return self.__fixIndex(i)

        return -1


    def GetLabels(self):
        """Returns the labels of all items in the list."""
        indices = map(self.__fixIndex, range(self.GetCount()))
        return [self.__listItems[i].label for i in indices]


    def GetData(self):
        """Returns the data associated with every item in the list."""
        indices = map(self.__fixIndex, range(self.GetCount()))
        return [self.__listItems[i].data for i in indices]


    def SetItemLabel(self, n, s):
        """Sets the label of the item at index ``n`` to the string ``s``.

        :arg n: Index of the item.
        :arg s: New label for the item.
        """

        if n < 0 or n >= self.GetCount():
            raise IndexError('Index {} is out of bounds'.format(n))

        n = self.__fixIndex(n)

        self.__listItems[n].labelWidget.SetLabel(s)
        self.__listItems[n].label = s


    def GetItemLabel(self, n):
        """Returns the label of the item at index ``n``.

        :arg n: Index of the item.
        """
        if n < 0 or n >= self.GetCount():
            raise IndexError('Index {} is out of bounds'.format(n))

        n = self.__fixIndex(n)

        return self.__listItems[n].label


    def SetItemWidget(self, n, widget=None):
        """Sets the widget to be displayed alongside the item at index ``n``.

        If ``widget`` is set to ``None``, any existing widget associated
        with the item is destroyed.
        """

        item  = self.__listItems[self.__fixIndex(n)]
        sizer = item.container.GetSizer()

        if item.extraWidget is not None:
            sizer.Detach(item.extraWidget)
            item.extraWidget.Destroy()
            item.extraWidget = None

        if widget is not None:
            item.extraWidget = widget
            widget.Reparent(item.container)
            sizer.Insert(0, widget)

        self.__updateScrollbar()


    def GetItemWidget(self, i):
        """Returns the widget for the item at index ``i``, or ``None``,
        if the widget hasn't been set.
        """

        return self.__listItems[i].extraWidget

    def SetItemTooltip(self, n, tooltip=None):
        """Sets the tooltip associated with the item at index ``n``."""
        n = self.__fixIndex(n)
        self.__listItems[n].tooltip = tooltip


    def GetItemTooltip(self, n):
        """Returns the tooltip associated with the item at index ``n``."""
        n = self.__fixIndex(n)
        return self.__listItems[n].tooltip


    def SetItemData(self, n, data=None):
        """Sets the data associated with the item at index ``n``."""
        n = self.__fixIndex(n)
        self.__listItems[n].data = data


    def GetItemData(self, n):
        """Returns the data associated with the item at index ``n``."""
        n = self.__fixIndex(n)
        return self.__listItems[n].data


    def SetItemForegroundColour(self,
                                n,
                                defaultColour=None,
                                selectedColour=None):
        """Sets the foreground colour of the item at index ``n``."""

        if defaultColour is None:
            defaultColour  = EditableListBox._defaultFG
            selectedColour = EditableListBox._selectedFG

        if selectedColour is None:
            selectedColour = defaultColour

        item = self.__listItems[self.__fixIndex(n)]

        item.defaultFGColour  = defaultColour
        item.selectedFGColour = selectedColour

        self.SetSelection(self.__fixIndex(self.__selection))


    def SetItemBackgroundColour(self,
                                n,
                                defaultColour=None,
                                selectedColour=None):
        """Sets the background colour of the item at index ``n``."""

        if defaultColour is None:
            defaultColour  = EditableListBox._defaultBG
            selectedColour = EditableListBox._selectedBG

        if selectedColour is None:
            selectedColour = defaultColour

        item = self.__listItems[self.__fixIndex(n)]

        item.defaultBGColour  = defaultColour
        item.selectedBGColour = selectedColour

        self.SetSelection(self.__fixIndex(self.__selection))


    def SetItemFont(self, n, font):
        """Sets the font for the item label at index ``n``."""
        li   = self.__listItems[self.__fixIndex(n)]
        li.labelWidget.SetFont(font)


    def GetItemFont(self, n):
        """Returns the font for the item label at index ``n``."""
        li = self.__listItems[self.__fixIndex(n)]
        return li.labelWidget.GetFont()


    def Enable(self, enable=True):
        """Enables/disables this ``EditableListBox`` and all of its children.
        """
        wx.Panel.Enable(self, enable)
        for item in self.__listItems:
            item.container.Enable(enable)


    def Disable(self):
        """Equivalent to ``Enable(False)``. """
        self.Enable(False)


    def ApplyFilter(self, filterStr=None, ignoreCase=False):
        """Hides any items for which the label does not contain the given
        ``filterStr``.

        To clear the filter (and hence show all items), pass in
        ``filterStr=None``.
        """

        if filterStr is None:
            filterStr = ''

        filterStr = filterStr.strip().lower()

        for item in self.__listItems:
            item.hidden = filterStr not in item.label.lower()

        self.__updateScrollbar()
        self.__drawList()


    def __getSelection(self, fix=False):
        """Returns a 3-tuple containing the (uncorrected) index, label,
        and associated client data of the currently selected list item,
        or (None, None, None) if no item is selected.
        """

        idx   = self.__selection
        label = None
        data  = None

        if idx == wx.NOT_FOUND:
            idx = None
        else:
            label = self.__listItems[idx].label
            data  = self.__listItems[idx].data

        if fix:
            idx = self.__fixIndex(idx)

        return idx, label, data


    def __itemClicked(self, ev=None, widget=None):
        """Called when an item in the list is clicked. Selects the item
        and posts an :data:`EVT_ELB_SELECT_EVENT`.

        This method may be called programmatically, by explicitly passing
        in the target ``widget``.  This functionality is used by the
        :meth:`__onKeyboard` event.

        :arg ev:     A :class:`wx.MouseEvent`.
        :arg widget: The widget on which to simulate a mouse click. Must
                     be provided when called programmatically.
        """

        # Give focus to the top level  panel,
        # otherwise it will not receive char events
        self.SetFocusIgnoringChildren()

        if ev is not None:
            widget = ev.GetEventObject()

        itemIdx = -1

        for i, listItem in enumerate(self.__listItems):
            if widget in (listItem.labelWidget, listItem.container):
                itemIdx = i
                break

        if itemIdx == -1:
            return

        self.SetSelection(self.__fixIndex(itemIdx))

        idx, label, data = self.__getSelection(True)

        log.debug('ListSelectEvent (idx: {}; label: {})'.format(idx, label))

        ev = ListSelectEvent(idx=idx, label=label, data=data)
        wx.PostEvent(self, ev)


    def __moveItem(self, offset):
        """Called when the *move up* or *move down* buttons are pushed.

        Moves the selected item by the specified offset and posts an
        :data:`EVT_ELB_MOVE_EVENT`, unless it doesn't make sense to do the
        move.
        """

        oldIdx, label, data = self.__getSelection()

        if oldIdx is None: return

        newIdx = oldIdx + offset

        # the selected item is at the top/bottom of the list.
        if oldIdx < 0 or oldIdx >= self.GetCount(): return
        if newIdx < 0 or newIdx >= self.GetCount(): return

        widget = self.__listSizer.GetItem(oldIdx).GetWindow()

        self.__listItems.insert(newIdx, self.__listItems.pop(oldIdx))

        self.__listSizer.Detach(oldIdx)
        self.__listSizer.Insert(newIdx, widget, flag=wx.EXPAND)

        oldIdx = self.__fixIndex(oldIdx)
        newIdx = self.__fixIndex(newIdx)

        self.SetSelection(newIdx)

        self.__listSizer.Layout()

        log.debug('ListMoveEvent (oldIdx: {}; newIdx: {}; label: {})'.format(
            oldIdx, newIdx, label))

        ev = ListMoveEvent(
            oldIdx=oldIdx, newIdx=newIdx, label=label, data=data)
        wx.PostEvent(self, ev)


    def __moveItemDown(self, ev):
        """Called when the *move down* button is pushed. Calls the
        :meth:`__moveItem` method.
        """
        self.__moveItem(1)


    def __moveItemUp(self, ev):
        """Called when the *move up* button is pushed. Calls the
        :meth:`__moveItem` method.
        """
        self.__moveItem(-1)


    def __addItem(self, ev):
        """Called when the *add item* button is pushed.

        Does nothing but post an :data:`EVT_ELB_ADD_EVENT` - it is up to a
        registered handler to implement the functionality of adding an item to
        the list.
        """

        idx, label, data = self.__getSelection(True)

        log.debug('ListAddEvent (idx: {}; label: {})'.format(idx, label))

        ev = ListAddEvent(idx=idx, label=label, data=data)

        wx.PostEvent(self, ev)


    def __removeItem(self, ev):
        """Called when the *remove item* button is pushed.

        Posts an :data:`EVT_ELB_REMOVE_EVENT` and removes the
        selected item from the list.

        Event listeners may call ``Veto()`` on the event object to cancel
        the removal.
        """

        idx, label, data = self.__getSelection(True)

        if idx is None: return

        ev = ListRemoveEvent(idx=idx, label=label, data=data)

        log.debug('ListRemoveEvent (idx: {}; label: {})'.format(
            idx, label))

        # We use ProcessEvent instead of wx.PostEvent,
        # because the latter processes the event
        # asynchronously, and we need to check whether
        # the event handler vetoed the event.
        self.GetEventHandler().ProcessEvent(ev)

        if ev.GetVeto():
            log.debug('ListRemoveEvent vetoed (idx: {}; label: {})'.format(
                idx, label))
            return

        self.Delete(idx)

        if self.GetCount() > 0:
            if idx == self.GetCount():
                self.SetSelection(idx - 1)
            else:
                self.SetSelection(idx)


    def __onEdit(self, ev, listItem):
        """Called when an item is double clicked.

        This method is only called if the :data:`ELB_EDITABLE` style flag is
        set.

        Creates and displays a :class:`wx.TextCtrl` allowing the user to edit
        the item label. A :class:`ListEditEvent` is posted every time the text
        changes.
        """
        idx      = self.__listItems.index(listItem)
        idx      = self.__fixIndex(idx)

        sizer    = listItem.container.GetSizer()
        editCtrl = wx.TextCtrl(listItem.container, style=wx.TE_PROCESS_ENTER)

        editCtrl.SetValue(listItem.label)

        # Listens to key presses. The edit is
        # cancelled if the escape key is pressed.
        def onKey(ev):
            ev.Skip()
            key = ev.GetKeyCode()
            if key == wx.WXK_ESCAPE: onFinish()

        # Destroyes the textctrl, and re-shows the item label.
        def onFinish(ev=None):

            if ev is not None:
                ev.Skip()

            def _onFinish():
                sizer.Detach(editCtrl)
                editCtrl.Destroy()
                sizer.Show(listItem.labelWidget, True)
                sizer.Layout()
                self.Refresh()

            wx.CallAfter(_onFinish)

        # Sets the list item label to the new
        # value, and posts a ListEditEvent.
        def onText(ev):
            oldLabel       = listItem.labelWidget.GetLabel()
            newLabel       = editCtrl.GetValue()
            listItem.label = newLabel
            listItem.labelWidget.SetLabel(newLabel)

            log.debug('ListEditEvent (idx: {}, oldLabel: {}, newLabel: {})'
                      .format(idx, oldLabel, newLabel))

            ev = ListEditEvent(idx=idx, label=newLabel, data=listItem.data)
            wx.PostEvent(self, ev)

        editCtrl.Bind(wx.EVT_TEXT,       onText)
        editCtrl.Bind(wx.EVT_KEY_DOWN,   onKey)
        editCtrl.Bind(wx.EVT_TEXT_ENTER, onFinish)
        editCtrl.Bind(wx.EVT_KILL_FOCUS, onFinish)

        if self.__widgetOnRight:
            sizer.Insert(0, editCtrl, flag=wx.EXPAND, proportion=1)
        else:
            sizer.Insert(2, editCtrl, flag=wx.EXPAND, proportion=1)

        sizer.Show(listItem.labelWidget, False)
        sizer.Layout()

        editCtrl.SetFocus()


    def __onDoubleClick(self, ev, listItem):
        """Called when an item is double clicked. See the :data:`ELB_EDITABLE`
        style.

        This method is only called if the :data:`ELB_EDITABLE` style flag is
        not set.

        Posts a :class:`ListDblClickEvent`.
        """

        idx = self.__listItems.index(listItem)
        idx = self.__fixIndex(idx)

        ev = ListDblClickEvent(idx=idx,
                               label=listItem.label,
                               data=listItem.data)
        wx.PostEvent(self, ev)


    def __updateMoveButtons(self):
        if self.__moveSupport:
            self.__upButton  .Enable((self.__selection != wx.NOT_FOUND) and
                                     (self.__selection != 0))
            self.__downButton.Enable((self.__selection != wx.NOT_FOUND) and
                                     (self.__selection != self.GetCount() - 1))


class _ListItem(object):
    """Internal class used to represent items in the list."""

    def __init__(self,
                 label,
                 data,
                 tooltip,
                 labelWidget,
                 container,
                 defaultFGColour,
                 selectedFGColour,
                 defaultBGColour,
                 selectedBGColour,
                 extraWidget=None):

        """Create a _ListItem object.

        :param str label:        The item label which will be displayed.

        :param data:             User data associated with the item.

        :param str tooltip:      A tooltip to be displayed when the mouse
                                 is moved over the item.

        :param labelWidget:      The :mod:`wx` object which represents the
                                 list item.

        :param container:        The :mod:`wx` object used as a container for
                                 the ``widget``.

        :param defaultFGColour:  Foreground colour to use when the item is
                                 not selected.

        :param selectedFGColour: Foreground colour to use when the item is
                                 selected.

        :param defaultBGColour:  Background colour to use when the item is
                                 not selected.

        :param selectedBGColour: Background colour to use when the item is
                                 selected.

        :param extraWidget:      A user-settable widget to be displayed
                                 alongside this item.
        """
        self.label            = label
        self.data             = data
        self.labelWidget      = labelWidget
        self.container        = container
        self.tooltip          = tooltip
        self.defaultFGColour  = defaultFGColour
        self.selectedFGColour = selectedFGColour
        self.defaultBGColour  = defaultBGColour
        self.selectedBGColour = selectedBGColour
        self.extraWidget      = extraWidget
        self.hidden           = False


_ListSelectEvent,   _EVT_ELB_SELECT_EVENT   = wxevent.NewEvent()
_ListAddEvent,      _EVT_ELB_ADD_EVENT      = wxevent.NewEvent()
_ListRemoveEvent,   _EVT_ELB_REMOVE_EVENT   = wxevent.NewEvent()
_ListMoveEvent,     _EVT_ELB_MOVE_EVENT     = wxevent.NewEvent()
_ListEditEvent,     _EVT_ELB_EDIT_EVENT     = wxevent.NewEvent()
_ListDblClickEvent, _EVT_ELB_DBLCLICK_EVENT = wxevent.NewEvent()


EVT_ELB_SELECT_EVENT = _EVT_ELB_SELECT_EVENT
"""Identifier for the :data:`ListSelectEvent` event."""


EVT_ELB_ADD_EVENT = _EVT_ELB_ADD_EVENT
"""Identifier for the :data:`ListAddEvent` event."""


EVT_ELB_REMOVE_EVENT = _EVT_ELB_REMOVE_EVENT
"""Identifier for the :data:`ListRemoveEvent` event."""


EVT_ELB_MOVE_EVENT = _EVT_ELB_MOVE_EVENT
"""Identifier for the :data:`ListMoveEvent` event."""


EVT_ELB_EDIT_EVENT = _EVT_ELB_EDIT_EVENT
"""Identifier for the :data:`ListEditEvent` event."""


EVT_ELB_DBLCLICK_EVENT = _EVT_ELB_DBLCLICK_EVENT


ListSelectEvent = _ListSelectEvent
"""Event emitted when an item is selected. A ``ListSelectEvent``
has the following attributes (all are set to ``None`` if no
item was selected):

- ``idx``:   Index of selected item
- ``label``: Label of selected item
- ``data``:  Client data associated with selected item
"""


ListAddEvent = _ListAddEvent
"""Event emitted when the 'add item' button is pushed. It is
up to a listener of this event to actually add a new item
to the list. A ``ListAddEvent`` has the following attributes
(all are set to ``None`` if no item was selected):

- ``idx``:   Index of selected item
- ``label``: Label of selected item
- ``data``:  Client data associated with selected item
"""


ListRemoveEvent = _ListRemoveEvent
"""Event emitted when the 'remove item' button is pushed. A
``ListRemoveEvent`` has the following attributes:

- ``idx``:   Index of removed item
- ``label``: Label of removed item
- ``data``:  Client data associated with removed item

An event handler can call ``ListRemoveEvent.Veto()`` to cancel
the item removal.
"""

def Veto(self):
    self._vetoed = True

def GetVeto(self):
    return getattr(self, '_vetoed', False)


ListRemoveEvent.Veto    = Veto
ListRemoveEvent.GetVeto = GetVeto


ListMoveEvent = _ListMoveEvent
"""Event emitted when one of the 'move up'/'move down'
buttons is pushed. A ``ListMoveEvent`` has the following
attributes:

- ``oldIdx``: Index of item before move
- ``newIdx``: Index of item after move
- ``label``:  Label of moved item
- ``data``:   Client data associated with moved item
"""


ListEditEvent = _ListEditEvent
"""Event emitted when a list item is edited by the user (see the
:data:`ELB_EDITABLE` style). A ``ListEditEvent`` has the following
attributes:

  - ``idx``:   Index of edited item
  - ``label``: New label of edited item
  - ``data``:  Client data associated with edited item.
"""

ListDblClickEvent = _ListDblClickEvent
"""Event emitted when a list item is double-clicked onthe user (see
the :data:`ELB_EDITABLE` style). A ``ListDblClickEvent`` has the
following attributes:

  - ``idx``:   Index of clicked item
  - ``label``: Label of clicked item
  - ``data``:  Client data associated with clicked item.
"""


ELB_NO_SCROLL = 1
"""If enabled, there will be no scrollbar. """


ELB_NO_ADD    = 2
"""If enabled, there will be no *add item* button."""


ELB_NO_REMOVE = 4
"""If enabled, there will be no *remove item* button."""


ELB_NO_MOVE   = 8
"""If enabled there will be no *move item up* or *move item down* buttons. """


ELB_REVERSE   = 16
"""If enabled, the first item in the list (index 0) will be shown at the
bottom, and the last item at the top.
"""


ELB_TOOLTIP   = 32
"""If enabled, list items will be replaced with a tooltip on mouse-over. If
disabled, a regular tooltip is shown.
"""


ELB_EDITABLE = 64
"""If enabled, double clicking a list item will allow the user to edit the
item value.

If this style is disabled, the :attr:`EVT_ELB_DBLCLICK_EVENT` will not
be generated.
"""

ELB_NO_LABELS = 128
"""If enabled, item labels are not shown - this is intended for lists which
are to consist solely of widgets (see the ``extraWidget`` parameter to the
:meth:`Insert` method). This style flag will negate the :data:`ELB_EDITABLE`
flag.
"""


ELB_WIDGET_RIGHT = 256
"""If enabled, item widgets are shown to the right of the item label.
Otherwise (by default) item widgets are shown to the left.
"""


ELB_TOOLTIP_DOWN = 512
"""If enabled, when the left mouse button is clicked and held down on a list
item, the item label is replaced with its tooltip while the mouse is held down.
This style is ignored if the :data:`ELB_TOOLTIP` style is active.
"""


ELB_SCROLL_BUTTONS = 1024
"""If enabled, and :data:`ELB_NO_SCROLL` is not enabled, up/down buttons are
added above/below the list, which allow the user to scroll up/down.
"""



#!/usr/bin/env python
#
# imagepanel.py - A panel for displaying a wx.Image.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ImagePanel` class, for displaying a
:class:`wx.Image`.
"""


import wx


class ImagePanel(wx.Panel):
    """A :class:`wx.Panel` which may be used to display a resizeable
    :class:`wx.Image`. The image is scaled to the size of the panel.
    """

    def __init__(self,
                 parent,
                 image=None,
                 preserveAspect=False):
        """Create an ``ImagePanel``.

        If the ``image`` is not passed in here, it can be set later with the
        :meth:`SetImage` method.

        :arg parent:         The :mod:`wx` parent object.

        :arg image:          The :class:`wx.Image` object to display.

        :arg preserveAspect: Defaults to ``False``. If ``True``, the image
                             aspect ratio is preserved.
        """

        wx.Panel.__init__(self, parent)

        self.Bind(wx.EVT_PAINT, self.Draw)
        self.Bind(wx.EVT_SIZE,  self.__onSize)

        self.SetImage(image)

        self.__preserveAspect = preserveAspect


    def SetImage(self, image):
        """Set the image that is displayed on this ``ImagePanel``.

        :arg image: The :class:`wx.Image` object to display.
        """
        self.__image = image

        self.Refresh()


    def __onSize(self, ev):
        """Redraw this panel when it is sized, so the image is scaled
        appropriately - see the :meth:`Draw` method.
        """
        self.Refresh()
        ev.Skip()


    def Draw(self, ev=None):
        """Draws this ``ImagePanel``. The image is scaled to the current panel
        size.
        """

        self.ClearBackground()

        if self.__image is None:
            return

        if ev is None: dc = wx.ClientDC(self)
        else:          dc = wx.PaintDC( self)

        if not dc.IsOk():
            return

        dwidth, dheight = dc.GetSize().Get()

        if dwidth == 0 or dheight == 0:
            return

        if self.__preserveAspect:
            iwidth, iheight = self.__image.GetSize().Get()
            iratio          = float(iwidth) / iheight
            dratio          = float(dwidth) / dheight

            # canvas is too wide - reduce
            # the display image width
            if dratio > iratio:
                dwidth = dheight / iratio

            # canvas is too tall - reduce
            # the display image height
            elif dratio < iratio:
                dheight = dwidth * iratio

        bitmap = self.__image.Scale(dwidth, dheight).ConvertToBitmap()

        dc.DrawBitmap(bitmap, 0, 0, False)



#!/usr/bin/env python
#
# test_isalive.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import wx

import fsleyes_widgets as fw

from . import run_with_wx, realYield


def test_isalive():
    run_with_wx(_test_isalive)

def _test_isalive():
    frame   = wx.GetApp().GetTopWindow()
    sizer   = wx.BoxSizer(wx.VERTICAL)
    panel   = wx.Panel(frame)
    child1  = wx.Panel(panel)
    child2  = wx.Panel(panel)

    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)

    sizer = wx.BoxSizer(wx.VERTICAL)
    sizer.Add(child1, flag=wx.EXPAND, proportion=1)
    sizer.Add(child2, flag=wx.EXPAND, proportion=1)
    panel.SetSizer(sizer)

    assert fw.isalive(child1)
    assert fw.isalive(child2)

    sizer.Detach(child1)
    child1.Destroy()

    sizer.Layout()
    frame.Refresh()

    realYield(100)

    assert not fw.isalive(child1)
    assert     fw.isalive(child2)


def test_wxversion():
    run_with_wx(_test_wxversion)
def _test_wxversion():
    fw.wxversion()



#!/usr/bin/env python
#
# bitmapradio.py - A radio control with bitmap buttons.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`BitmapRadioBox` class, a radio control
which uses bitmap toggle buttons.
"""

import wx
import wx.lib.newevent as wxevent

from . import bitmaptoggle


# TODO Button tooltips


class BitmapRadioBox(wx.Panel):
    """A radio control which displays a collection of :class:`wx.ToggleButton`
    controls, each of which displays an image.

    Each of these buttons corresponds to a mutually exclusive option - when
    the user clicks on a button, it is toggled *on*, and all of the others are
    toggled *off*.

    For example, here is a ``BitmapRadioBox`` which allows the user to switch
    between *view* mode and *edit* mode:

     .. image:: images/bitmapradiobox.png
        :scale: 50%
        :align: center

    When the user pushes a button, a :data:`EVT_BITMAP_RADIO_EVENT` is
    generated.
    """

    def __init__(self, parent, style=None):
        """Create a ``BitmapRadioBox``.

        :arg parent: A parent window.

        :arg style:  Either ``wx.HORIZONTAL`` (the default) or ``wx.VERTICAL``,
                     to control the button layout direction.
        """
        wx.Panel.__init__(self, parent)

        if style is None:
            style = wx.HORIZONTAL

        if style & wx.VERTICAL: szorient = wx.VERTICAL
        else:                   szorient = wx.HORIZONTAL

        self.__selection  = -1
        self.__buttons    = []
        self.__clientData = []
        self.__sizer      = wx.BoxSizer(szorient)

        self.SetSizer(self.__sizer)


    def AddChoice(self, selectedBmp, unselectedBmp=None, clientData=None):
        """Add a button to this ``BitmapRadioBox``.

        :arg selectedBmp:   A :class:`wx.Bitmap` to display on the button when
                            it is selected.

        :arg unselectedBmp: Optional. A :class:`wx.Bitmap` to display on the
                            button when it is not selected.

        :arg clientData:    Arbitrary data which is associated with the choice.
        """

        style  = wx.BU_EXACTFIT | wx.ALIGN_CENTRE | wx.BU_NOTEXT
        button = bitmaptoggle.BitmapToggleButton(self,
                                                 trueBmp=selectedBmp,
                                                 falseBmp=unselectedBmp,
                                                 style=style)

        self.__buttons   .append(button)
        self.__clientData.append(clientData)

        self.__sizer.Add(button, flag=wx.EXPAND)
        self.Layout()

        button.Bind(bitmaptoggle.EVT_BITMAP_TOGGLE, self.__onButton)

        if self.__selection == -1:
            self.SetSelection(0)


    def Clear(self):
        """Remove all buttons from this ``BitmapRadioBox``."""

        self.__sizer.Clear(True)
        self.__selection  = -1
        self.__buttons    = []
        self.__clientData = []


    def EnableChoice(self, index, enable=True):
        """Enable or disable the button at the specified index. """
        self.__buttons[index].Enable(enable)


    def DisableChoice(self, index=None):
        """Disable the button at the specified index. """
        self.EnableChoice(index, False)


    def Set(self, bitmaps, clientData=None):
        """Set all buttons at once.

        :arg bitmaps:    A list of :class:`wx.Bitmap` objects.

        :arg clientData: A list of arbitrary data to associate with each
                         choice.
        """

        if clientData is None:
            clientData = [None] * len(bitmaps)

        self.Clear()
        for bmp, cd in zip(bitmaps, clientData):
            self.AddChoice(bmp, clientData=cd)


    def GetSelection(self):
        """Returns the index of the curently selected choice."""
        return self.__selection


    def SetSelection(self, index):
        """Sets the current selection."""

        if index < 0 or index >= len(self.__buttons):
            raise IndexError('Invalid index {}'.format(index))

        self.__selection = index

        for i, button in enumerate(self.__buttons):

            if i == index: button.SetValue(True)
            else:          button.SetValue(False)


    def __onButton(self, ev):
        """Called when a button is pushed. Updates the selection,
        and emits a :data:`BitmapRadioEvent`.
        """

        idx  = self.__buttons.index(ev.GetEventObject())
        data = self.__clientData[idx]

        self.SetSelection(idx)

        wx.PostEvent(self, BitmapRadioEvent(index=idx, clientData=data))


_BitampRadioEvent, _EVT_BITMAP_RADIO_EVENT = wxevent.NewEvent()


EVT_BITMAP_RADIO_EVENT = _EVT_BITMAP_RADIO_EVENT
"""Identifier for the :data:`BitmapRadioEvent`. """


BitmapRadioEvent = _BitampRadioEvent
"""Event emitted when the user changes the radio selection. Contains
the following attributes:

  - ``index``:      The index of the new selection.
  - ``clientData``: Client data associated with the new selection.
"""



#!/usr/bin/env python
#
# test_bitmaptoggle.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os.path as op

import wx

import fsleyes_widgets.bitmaptoggle as bmptoggle

from . import run_with_wx, simclick

datadir = op.join(op.dirname(__file__), 'testdata', 'bitmapbuttons')


def test_Create():
    run_with_wx(_test_Create)
def _test_Create():

    frame = wx.GetApp().GetTopWindow()


    falseicon = op.join(datadir, 'false.png')
    trueicon  = op.join(datadir, 'true.png')

    falseicon = wx.Bitmap(falseicon, wx.BITMAP_TYPE_PNG)
    trueicon  = wx.Bitmap(trueicon,  wx.BITMAP_TYPE_PNG)

    btn = bmptoggle.BitmapToggleButton(frame)
    btn = bmptoggle.BitmapToggleButton(frame,
                                       trueBmp=trueicon)
    btn = bmptoggle.BitmapToggleButton(frame,
                                       trueBmp=trueicon,
                                       falseBmp=falseicon)

    btn = bmptoggle.BitmapToggleButton(frame)
    btn.SetBitmap(trueicon)
    btn.SetBitmap(trueicon, falseicon)


def test_SetGet():
    run_with_wx(_test_SetGet)
def _test_SetGet():

    frame = wx.GetApp().GetTopWindow()
    btn = bmptoggle.BitmapToggleButton(frame)

    btn.SetValue(True)
    assert btn.GetValue()

    btn.SetValue(False)
    assert not btn.GetValue()


def test_Toggle():
    run_with_wx(_test_Toggle)
def _test_Toggle():
    falseicon = op.join(datadir, 'false.png')
    trueicon  = op.join(datadir, 'true.png')

    falseicon = wx.Bitmap(falseicon, wx.BITMAP_TYPE_PNG)
    trueicon  = wx.Bitmap(trueicon,  wx.BITMAP_TYPE_PNG)

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    btn   = bmptoggle.BitmapToggleButton(frame)
    btn.SetBitmap(trueicon, falseicon)

    assert not btn.GetValue()
    simclick(sim, btn)
    assert btn.GetValue()


def test_Event():
    run_with_wx(_test_Event)
def _test_Event():
    trueicon = op.join(datadir, 'true.png')
    trueicon = wx.Bitmap(trueicon,  wx.BITMAP_TYPE_PNG)

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    btn   = bmptoggle.BitmapToggleButton(frame)
    btn.SetBitmap(trueicon)

    result = [None]

    def handler(ev):
        result[0] = ev.value

    btn.Bind(bmptoggle.EVT_BITMAP_TOGGLE, handler)

    simclick(sim, btn)
    assert result[0]
    simclick(sim, btn)
    assert not result[0]



#!/usr/bin/env python
#
# test_placeholder_textctrl.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import wx

from . import run_with_wx

import fsleyes_widgets.placeholder_textctrl as ptc


def test_PlaceholderTextCtrl():
    run_with_wx(_test_PlaceholderTextCtrl)
def _test_PlaceholderTextCtrl():
    frame = wx.GetApp().GetTopWindow()
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    dummy = wx.TextCtrl(frame)
    ctrl  = ptc.PlaceholderTextCtrl(
        frame,
        placeholder='Enter text',
        placeholderColour=(200, 100, 100))
    sizer.Add(dummy, flag=wx.EXPAND)
    sizer.Add(ctrl,  flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()
    wx.Yield()

    assert ctrl.GetPlaceholder()       == 'Enter text'
    assert tuple(ctrl.GetPlaceholderColour()) == (200, 100, 100)
    assert ctrl.GetValue() == ''
    ctrl.SetValue('Bah, humbug')
    assert ctrl.GetValue() == 'Bah, humbug'
    ctrl.SetValue('')

    dummy.SetFocus()
    wx.Yield()
    ctrl.SetFocus()
    wx.Yield()
    dummy.SetFocus()
    wx.Yield()



#!/usr/bin/env python
#
# widgetgrid.py - A tabular grid of widgets.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WidgetGrid` class, which can display a
tabular grid of arbitrary widgets.
"""


import logging

import wx
import wx.lib.newevent as wxevent


log = logging.getLogger(__name__)


class WidgetGrid(wx.ScrolledWindow):
    """A scrollable panel which displays a tabular grid of widgets.  A
    ``WidgetGrid`` looks something like this:

      .. image:: images/widgetgrid.png
        :scale: 50%
        :align: center


    The most important methods are:

     .. autosummary::
        :nosignatures:

        GetGridSize
        SetGridSize
        Refresh
        DeleteRow
        InsertRow
        SetColours
        SetWidget
        SetText
        ClearGrid


    *Labels*

    .. autosummary::
       :nosignatures:

       ShowRowLabels
       ShowColLabels
       SetRowLabel
       SetColLabel


    *Selections*

    .. autosummary::
       :nosignatures:

       GetSelection
       SetSelection


    *Styles*


    The ``WidgetGrid`` supports the following styles:

    =============================== ================================
    ``wx.HSCROLL``                  Use a horizontal scrollbar.
    ``wx.VSCROLL``                  Use a vertical scrollbar.
    :data:`WG_SELECTABLE_CELLS`     Individual cells are selectable.
    :data:`WG_SELECTABLE_ROWS`      Rows are selectable.
    :data:`WG_SELECTABLE_COLUMN`    Columns are selectable.
    :data:`WG_KEY_NAVIGATION`       The keyboard can be used for
                                    navigation
    =============================== ================================


    The ``*_SELECTABLE_*`` styles are mutualliy exclusive; their precedence
    is equivalent to their order in the above table. By default, the arrow
    keys are used for keyboard navigation, but these are customisable via
    the :meth:`SetNavKeys` method.


    *Events*

    The following events may be emitted by a ``WidgetGrid``:

    .. autosummary::
       :nosignatures:

       :data:`WidgetGridSelectEvent`
    """


    _defaultBorderColour = '#000000'
    """The colour of border a border which is shown around every cell in the
    grid.
    """


    _defaultOddColour    = '#ffffff'
    """Background colour for cells on odd rows."""


    _defaultEvenColour   = '#eeeeee'
    """Background colour for cells on even rows."""


    _defaultLabelColour  = '#dddddd'
    """Background colour for row and column labels."""


    _defaultSelectedColour = '#cdcdff'
    """Background colour for selected cells. """


    def __init__(self, parent, style=None):
        """Create a ``WidgetGrid``.

        :arg parent: The :mod:`wx` parent object.
        :arg style:  Style flags  - can be a combination of ``wx.HSCROLL``,
                     ``wx.VSCROLL``, :data:`WG_SELECTABLE_CELLS`,
                     :data:`WG_SELECTABLE_ROWS`, and
                     :data:`WG_SELECTABLE_COLUMNS`.
        """

        if style is None:
            style = wx.HSCROLL | wx.VSCROLL

        self.__hscroll = style & wx.HSCROLL
        self.__vscroll = style & wx.VSCROLL
        self.__keynav  = style & WG_KEY_NAVIGATION

        if   style & WG_SELECTABLE_CELLS:   self.__selectable = 'cells'
        elif style & WG_SELECTABLE_ROWS:    self.__selectable = 'rows'
        elif style & WG_SELECTABLE_COLUMNS: self.__selectable = 'columns'
        else:
            self.__keynav     = False
            self.__selectable = None

        wx.ScrolledWindow.__init__(self, parent, style=wx.WANTS_CHARS)

        hrate = 1 if self.__hscroll else 0
        vrate = 1 if self.__vscroll else 0
        self.SetScrollRate(hrate, vrate)

        self.__gridPanel = wx.Panel(self, style=wx.WANTS_CHARS)

        self.__sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.__sizer)
        self.__sizer.Add(self.__gridPanel, flag=wx.EXPAND)

        # The __widgets array contains wx.Panel
        # objects which are used as containers
        # for all widgets added to the grid.
        # The __widgetRefs array contains the
        # actual widget objects that were passed
        # to the SetWidget method.
        self.__gridSizer      = None
        self.__nrows          = 0
        self.__ncols          = 0
        self.__widgets        = []
        self.__widgetRefs     = []
        self.__rowLabels      = []
        self.__colLabels      = []
        self.__selected       = None

        self.__showRowLabels  = False
        self.__showColLabels  = False
        self.__borderColour   = WidgetGrid._defaultBorderColour
        self.__labelColour    = WidgetGrid._defaultLabelColour
        self.__oddColour      = WidgetGrid._defaultOddColour
        self.__evenColour     = WidgetGrid._defaultEvenColour
        self.__selectedColour = WidgetGrid._defaultSelectedColour
        self.__upKey          = wx.WXK_UP
        self.__downKey        = wx.WXK_DOWN
        self.__leftKey        = wx.WXK_LEFT
        self.__rightKey       = wx.WXK_RIGHT

        self.Bind(wx.EVT_SIZE, self.__onResize)

        if self.__keynav:
            self.Bind(wx.EVT_KEY_DOWN, self.__onKeyboard)

        if self.__selectable:
            # A silly internal multi-level semaphore
            # used to ignore child focus events when
            # this WidgetGrid generated them.  See
            # the __selectCell method for more silly
            # comments.
            self.__ignoreFocus = 0
            self.Bind(wx.EVT_CHILD_FOCUS, self.__onChildFocus)


    def Refresh(self):
        """Redraws the contents of this ``WidgetGrid``. This method must be
        called after the contents of the grid are changed.
        """
        self.__refresh()


    def Hide(self):
        """Hides this ``WidgetGrid``. """
        self.Show(False)


    def Show(self, show=True):
        """Shows/hides this ``WidgetGrid``, and recursively does the same
        to all of its children.
        """

        def realShow(obj):
            if obj is self: wx.ScrolledWindow.Show(self, show)
            else:           obj.Show(show)

            for child in obj.GetChildren():
                realShow(child)

        realShow(self)


    def SetColours(self, **kwargs):
        """Set the colours used in this ``WidgetGrid``. The :meth:`Refresh`
        method must be called afterwards for this method to take effect.

        :arg border:   The cell border colour.

        :arg label:    Background colour for row and column labels.

        :arg odd:      Background colour for cells on odd rows.

        :arg even:     Background colour for cells on even rows.

        :arg selected: Background colour for selected cells.
        """

        border   = kwargs.get('border',   self)
        label    = kwargs.get('label',    self)
        odd      = kwargs.get('odd',      self)
        even     = kwargs.get('even',     self)
        selected = kwargs.get('selected', self)

        if border   is not self: self.__borderColour   = border
        if label    is not self: self.__labelColour    = label
        if odd      is not self: self.__oddColour      = odd
        if even     is not self: self.__evenColour     = even
        if selected is not self: self.__selectedColour = selected


    def SetNavKeys(self, **kwargs):
        """Set the keys used for keyboard navigation (if the
        :data:`WG_KEY_NAVIGATION` style is enabled). Setting an argument
        to ``None`` will disable navigation in that direction.

        :arg up:    Key to use for up navigation.
        :arg down:  Key to use for down navigation.
        :arg left:  Key to use for left navigation.
        :arg right: Key to use for right navigation.
        """

        up    = kwargs.get('up',    self)
        down  = kwargs.get('down',  self)
        left  = kwargs.get('left',  self)
        right = kwargs.get('right', self)

        self.__upKey    = up
        self.__downKey  = down
        self.__leftKey  = left
        self.__rightKey = right


    def __onResize(self, ev):
        """Called when this ``WidgetGrid`` is resized. Makes sure the
        scrollbars are up to date.
        """
        self.FitInside()


    def __reparent(self, widget, parent):
        """Convenience method which  re-parents the given widget. If
        ``widget`` is a :class:`wx.Sizer` the sizer children are re-parented.
        """
        if isinstance(widget, wx.Sizer):
            widget = [c.GetWindow() for c in widget.GetChildren()]
        else:
            widget = [widget]

        for w in widget:
            w.Reparent(parent)


    def __setBackgroundColour(self, widget, colour):
        """Convenience method which changes the background colour of the given
        widget. If ``widget`` is a :class:`wx.Sizer` the background colours of
        the sizer children is updated.
        """
        if isinstance(widget, wx.Sizer):
            widget = [c.GetWindow() for c in widget.GetChildren()]
        else:
            widget = [widget]

        for w in widget:
            w.SetBackgroundColour(colour)


    def __refresh(self):
        """Lays out and re-sizes the entire widget grid. """

        borderColour = self.__borderColour
        labelColour  = self.__labelColour
        oddColour    = self.__oddColour
        evenColour   = self.__evenColour

        if borderColour is None: borderColour = WidgetGrid._defaultBorderColour
        if labelColour  is None: labelColour  = WidgetGrid._defaultLabelColour
        if oddColour    is None: oddColour    = WidgetGrid._defaultOddColour
        if evenColour   is None: evenColour   = WidgetGrid._defaultEvenColour

        if self.__gridSizer is None:
            self.FitInside()
            self.Layout()
            return

        self.__gridPanel.SetBackgroundColour(borderColour)

        # Clear the sizer per-item, as the
        # wx.Sizer.Clear will destroy any
        # child sizers, and we don't want that
        for i in reversed(range(self.__gridSizer.GetItemCount())):
            self.__gridSizer.Detach(i)

        # empty cell in top left of grid
        self.__gridSizer.Add((-1, -1), flag=wx.EXPAND)

        # column labels
        for coli, (lblPanel, colLabel) in enumerate(self.__colLabels):

            # 1px border between every column
            if coli == self.__ncols - 1: flag = wx.TOP | wx.LEFT | wx.RIGHT
            else:                        flag = wx.TOP | wx.LEFT

            lblPanel.SetBackgroundColour(labelColour)
            colLabel.SetBackgroundColour(labelColour)

            self.__gridSizer.Add( lblPanel, border=1, flag=wx.EXPAND | flag)
            self.__gridSizer.Show(lblPanel, self.__showColLabels)

        # Rows
        for rowi in range(self.__nrows):

            lblPanel, rowLabel = self.__rowLabels[rowi]

            if rowi == self.__nrows - 1: flag = wx.TOP | wx.LEFT | wx.BOTTOM
            else:                        flag = wx.TOP | wx.LEFT

            lblPanel.SetBackgroundColour(labelColour)
            rowLabel.SetBackgroundColour(labelColour)

            self.__gridSizer.Add( lblPanel, border=1, flag=wx.EXPAND | flag)
            self.__gridSizer.Show(lblPanel, self.__showRowLabels)

            # Widgets
            for coli in range(self.__ncols):

                widget    = self.__widgetRefs[rowi][coli]
                container = self.__widgets[   rowi][coli]

                flag = wx.TOP | wx.LEFT

                if rowi == self.__nrows - 1: flag |= wx.BOTTOM
                if coli == self.__ncols - 1: flag |= wx.RIGHT

                self.__gridSizer.Add(container,
                                     flag=wx.EXPAND | flag,
                                     border=1,
                                     proportion=1)

                if rowi % 2: colour = oddColour
                else:        colour = evenColour

                self.__setBackgroundColour(container, colour)
                self.__setBackgroundColour(widget,    colour)

        if self.__selected is not None:
            row, col = self.__selected
            self.SetSelection(row, col)

        self.FitInside()
        self.Layout()


    def GetGridSize(self):
        """Returns the current grid size as a tuple containing ``(rows, cols)``.
        """
        return self.__nrows, self.__ncols


    def SetGridSize(self, nrows, ncols, growCols=None):
        """Set the size of the widdget grid. The :meth:`Refresh` method must
        be called afterwards for this method to take effect.

        :arg nrows:    Number of rows

        :arg ncols:    Number of columns

        :arg growCols: A sequence specifying which columns should be stretched
                       to fit.
        """

        if nrows < 0 or ncols < 0:
            raise ValueError('Invalid size ({}, {})'.format(nrows, ncols))

        # If the caller has not specified which columns
        # should stretch, then stretch them all so the
        # grid is sized to fit the available space.
        if growCols is None:
            growCols = range(ncols)

        self.ClearGrid()

        self.__nrows     = nrows
        self.__ncols     = ncols
        self.__gridSizer = wx.FlexGridSizer(nrows + 1, ncols + 1, 0, 0)

        self.__gridSizer.SetFlexibleDirection(wx.BOTH)

        for col in growCols:
            self.__gridSizer.AddGrowableCol(col + 1)

        self.__widgets    = [[None] * ncols for i in range(nrows)]
        self.__widgetRefs = [[None] * ncols for i in range(nrows)]
        self.__rowLabels  =  [None] * nrows
        self.__colLabels  =  [None] * ncols

        for rowi in range(nrows):
            for coli in range(ncols):
                self.__initCell(rowi, coli)

        for rowi in range(nrows): self.__initRowLabel(rowi)
        for coli in range(ncols): self.__initColLabel(coli)

        self.__gridPanel.SetSizer(self.__gridSizer)


    def __initCell(self, row, col):
        """Called by :meth:`SetGridSize` and :meth:`InsertRow`. Creates a
        placeholder ``wx.Panel`` at the specified cell.
        """
        placeholder = wx.Panel(self.__gridPanel)
        self.__widgets[   row][col] = placeholder
        self.__widgetRefs[row][col] = placeholder


    def __initRowLabel(self, row):
        """Called by :meth:`SetGridSize` and :meth:`InsertRow`. Creates a
        label widget at the specified row.
        """
        panel = wx.Panel(self.__gridPanel)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        lbl   = wx.StaticText(
            panel,
            style=wx.ALIGN_LEFT | wx.ALIGN_CENTER_VERTICAL)

        panel.SetSizer(sizer)
        sizer.Add(lbl, flag=wx.CENTRE)

        self.__initWidget(panel, row, -1)
        self.__initWidget(lbl,   row, -1)
        self.__rowLabels[row] = (panel, lbl)


    def __initColLabel(self, col):
        """Called by :meth:`SetGridSize`. Creates a label widget at the
        specified column
        """
        panel = wx.Panel(self.__gridPanel)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        lbl   = wx.StaticText(
            panel,
            style=wx.ALIGN_CENTRE_HORIZONTAL | wx.ALIGN_CENTRE_VERTICAL)

        panel.SetSizer(sizer)
        sizer.Add(lbl, flag=wx.CENTRE)

        self.__initWidget(panel, -1, col)
        self.__initWidget(lbl,   -1, col)
        self.__colLabels[col] = (panel, lbl)


    def GetRow(self, widget):
        """Returns the index of the row in which the given ``widget`` is
        located, or ``-1`` if it is not in the ``WidgetGrid``.
        """
        try:                   return widget._wg_row
        except AttributeError: return -1


    def GetColumn(self, widget):
        """Returns the index of the column in which the given ``widget`` is
        located, or ``-1`` if it is not in the ``WidgetGrid``.
        """
        try:                   return widget._wg_row
        except AttributeError: return -1


    def DeleteRow(self, row):
        """Removes the specified ``row`` from the grid, destroying all
        widgets on that row. This method does not need to be followed
        by a call to :meth:`Refresh`, but a call to ``Layout`` may be
        required.

        .. note:: Make sure you reparent any widgets that you do not want
                  destroyed before calling this method.
        """

        if self.__gridSizer is None:
            raise ValueError('No grid')

        if row < 0 or row >= self.__nrows:
            raise ValueError('Invalid row index {}'.format(row))

        log.debug('Deleting row {} (sizer indices {} - {})'.format(
            row,
            (row + 1) * (self.__ncols + 1),
            (row + 1) * (self.__ncols + 1) + self.__ncols))

        # Remove from the grid
        for col in reversed(range(self.__ncols + 1)):
            self.__gridSizer.Detach((row + 1) * (self.__ncols + 1) + col)

        # Update row/col references
        # on all widgets/labels
        for ri in range(row, self.__nrows):
            for ci in range(self.__ncols):
                self.__widgetRefs[ri][ci]._wg_row -= 1
                self.__widgets[   ri][ci]._wg_row -= 1

            self.__rowLabels[ri][0]._wg_row -= 1
            self.__rowLabels[ri][1]._wg_row -= 1

        # Destroy the widgets and the row label
        for widget in self.__widgets[row]:
            widget.Destroy()

        self.__rowLabels .pop(row)[0].Destroy()

        # Remove references to them
        self.__widgetRefs.pop(row)
        self.__widgets   .pop(row)

        # Update the grid
        self.__nrows -= 1
        self.__gridSizer.SetRows(self.__nrows + 1)

        # Update selected widget if necessary
        if self.__selected is not None:
            srow, scol = self.__selected
            if   srow == row:             self.__selected = None
            elif srow > row and srow > 0: self.__selected = (srow - 1, scol)

        self.FitInside()


    def InsertRow(self, row):
        """Inserts a new row into the ``WidgetGrid`` at the specified ``row``
        index. This method must be followed by a call to :meth:`Refresh`.
        """

        if self.__gridSizer is None:
            raise ValueError('No grid')

        if row < 0:
            raise ValueError('Invalid row index {}'.format(row))

        if row >= self.__nrows:
            row = self.__nrows

        log.debug('Inserting row at {}'.format(row))

        # Add empty label/cell
        # values for the new row
        self.__rowLabels .insert(row,  None)
        self.__widgets   .insert(row, [None] * self.__ncols)
        self.__widgetRefs.insert(row, [None] * self.__ncols)

        # Update the grid
        self.__nrows += 1
        self.__gridSizer.SetRows(self.__nrows + 1)

        # Initialise the contents
        # of the new row
        self.__initRowLabel(row)
        for col in range(self.__ncols):
            self.__initCell(row, col)

        # update row/col indices
        for ri in range(row + 1, self.__nrows):
            for ci in range(self.__ncols):
                self.__widgetRefs[ri][ci]._wg_row += 1
                self.__widgets[   ri][ci]._wg_row += 1

            self.__rowLabels[ri][0]._wg_row += 1
            self.__rowLabels[ri][1]._wg_row += 1

        # Update selected widget if necessary
        if self.__selected is not None:
            srow, scol = self.__selected

            if srow >= row and srow < self.__nrows - 1:
                self.__selected = (srow + 1, scol)


    def ClearGrid(self):
        """Removes and destroys all widgets from the grid, and sets the grid
        size to ``(0, 0)``. The :meth:`Refresh` method must be called
        afterwards for this method to take effect.
        """

        if self.__gridSizer is not None:
            self.__gridSizer.Clear(True)

        self.__gridSizer  = None
        self.__nrows      = 0
        self.__ncols      = 0
        self.__widgets    = []
        self.__widgetRefs = []
        self.__rowLabels  = []
        self.__colLabels  = []
        self.__selected   = None

        self.__gridPanel.SetSizer(None)


    def SetText(self, row, col, text):
        """Convenience method which creates a :class:`wx.StaticText` widget
        with the given text, and passes it to the :meth:`SetWidget` method.

        :arg row:  Row index.

        :arg col:  Column index.

        :arg text: Text to display.
        """

        txt = wx.StaticText(self.__gridPanel, label=text)
        self.SetWidget(row, col, txt)


    def GetWidget(self, row, col):
        """Returns the widget located at the specified row/column. """

        return self.__widgetRefs[row][col]


    def SetWidget(self, row, col, widget):
        """Adds the given widget to the grid. The :meth:`Refresh` method
        must be called afterwards for this method to take effect.

        The parent of the widget is changed to this ``WidgetGrid``.

         .. note:: The provided widget may alternately be a
                   :class:`wx.Sizer`. However, nested sizers, i.e. sizers
                   which contain other sizers, are not supported.

        :arg row:    Row index.

        :arg col:    Column index.

        :arg widget: The widget or sizer to add to the grid.

        Raises an :exc:`IndexError` if the specified grid location ``(row,
        col)`` is invalid.
        """
        if row <  0            or \
           col <  0            or \
           row >= self.__nrows or \
           col >= self.__ncols:
            raise IndexError('Grid location ({}, {}) out of bounds ({}, {})'.
                             format(row, col, self.__nrows, self.__ncols))

        # Embed the widget in a panel,
        # as Linux/GTK has trouble
        # changing the background colour
        # of some controls
        panel = wx.Panel(self.__gridPanel, style=wx.WANTS_CHARS)
        sizer = wx.BoxSizer(wx.HORIZONTAL)
        panel.SetSizer(sizer)

        self.__reparent(widget, panel)

        self.__initWidget(widget, row, col)
        self.__initWidget(panel,  row, col)

        sizer.Add(widget, flag=wx.EXPAND, proportion=1)

        if self.__widgets[row][col] is not None:
            self.__widgets[row][col].Destroy()

        self.__widgetRefs[row][col] = widget
        self.__widgets[   row][col] = panel


    def __initWidget(self, widget, row, col):
        """Called by the :meth:`AddWidget` method.

        Performs some initialisation on a widget which has just been added to
        the grid.

        :arg widget: The widget to initialise
        :arg row:    Row index of the widget in the grid.
        :arg col:    Column index of the widget in the grid.
        """

        def scroll(ev):
            posx, posy = self.GetViewStart()
            rotation   = ev.GetWheelRotation()

            if   rotation > 0: delta =  5
            elif rotation < 0: delta = -5
            else:              return

            if ev.GetWheelAxis() == wx.MOUSE_WHEEL_VERTICAL: posy -= delta
            else:                                            posx += delta

            self.Scroll(posx, posy)

        def initWidget(w):

            # Under Linux/GTK, we need to bind a mousewheel
            # listener to every child of the panel in order
            # for scrolling to work correctly. This is not
            # necessary under OSX/cocoa.
            if wx.Platform == '__WXGTK__':
                w.Bind(wx.EVT_MOUSEWHEEL, scroll)

            # Listen for mouse down events
            # if cells are selectable
            if self.__selectable and not w.AcceptsFocus():
                w.Bind(wx.EVT_LEFT_DOWN, self.__onLeftMouseDown)

            # Attach the row/column indices to the
            # widget - they are used in the
            # __onLeftDown and __onChildFocus methods
            w._wg_row = row
            w._wg_col = col

        if isinstance(widget, wx.Sizer):
            for c in widget.GetChildren():
                initWidget(c.GetWindow())
        else:
            initWidget(widget)


    def __selectCell(self, row, col):
        """Called by the :meth:`__onChildFocus` and :meth:`__onLeftMouseDown`
        methods. Selects the specified row/column, and generates an
        :data:`EVT_WG_SELECT` event.
        """

        if   self.__selectable == 'rows':    col = -1
        elif self.__selectable == 'columns': row = -1

        try:
            if not self.SetSelection(row, col):
                return

        except ValueError:
            return

        log.debug('Posting grid select event ({}, {})'.format(row, col))

        # This is a ridiculous workaround to a
        # ridiculous problem. Certain users of
        # the WidgetGrid focus a widget within
        # the grid on a select event. This
        # triggers a call to __onChildFocus.
        # Now, the logic in __onChildFocus
        # should be smart enough to not generate
        # another select event when the selected
        # cell hasn't changed. But under GTK it
        # seems that if keyboard/mouse events
        # occur fast enough, the order in which
        # event objects are passed becomes
        # unstable. This means that we can get
        # caught in a silly infinite focus
        # switching loop.
        #
        # To avoid this, I'm setting a flag here,
        # to tell the __onChildFocus method to
        # do nothing while any grid select event
        # handlers are running.
        self.__ignoreFocus += 1

        event = WidgetGridSelectEvent(row=row, col=col)
        event.SetEventObject(self)
        wx.PostEvent(self, event)

        def resetFocus():
            self.__ignoreFocus -= 1

        wx.CallAfter(resetFocus)


    def __onChildFocus(self, ev):
        """If this ``WidgetGrid`` is selectable, this method is called when a
        widget in the grid gains focus. Ensures that the containing cell is
        selected.
        """

        # We explicitly do not call ev.Skip
        # because otherwise the native wx code
        # will automatically scroll to show the
        # focused child, which potentially
        # interferes with application code. The
        # __selectCell method calls SetSelection
        # which then calls __scrollTo, which
        # makes sure that the selected cell is
        # visible.

        # See silliness in __selectCell
        if self.__ignoreFocus > 0:
            return

        gridWidget = ev.GetEventObject()

        row = None
        col = None

        # The event source may be a child of the
        # widget that was added to the grid, so we
        # search up the hierarchy to find the
        # parent that has row and column attributes
        while gridWidget is not None:

            if hasattr(gridWidget, '_wg_row'):
                row = gridWidget._wg_row
                col = gridWidget._wg_col
                break
            else:
                gridWidget = gridWidget.GetParent()

        if row is not None and col is not None:
            log.debug('Focus on cell {}'.format((row, col)))
            self.__selectCell(row, col)


    def __onLeftMouseDown(self, ev):
        """If this ``WidgetGrid`` is selectable, this method is called
        whenever an left mouse down event occurs on an item in the grid.
        """

        widget = ev.GetEventObject()
        row    = widget._wg_row
        col    = widget._wg_col

        log.debug('Left mouse down on cell {}'.format((row, col)))

        self.__selectCell(row, col)


    def __onKeyboard(self, ev):
        """If the :data:`WG_KEY_NAVIGATION` style is enabled, this method is
        called when the user pushes a key while this ``WidgetGrid`` has focus.
        It changes the currently selected cell, row, or column.
        """
        ev.ResumePropagation(wx.EVENT_PROPAGATE_MAX)

        key = ev.GetKeyCode()

        up    = self.__upKey
        down  = self.__downKey
        left  = self.__leftKey
        right = self.__rightKey

        log.debug('Keyboard event ({})'.format(key))

        if key not in (up, down, left, right):
            ev.Skip()
            return

        row, col = self.__selected

        if   key == up:    row -= 1
        elif key == down:  row += 1
        elif key == left:  col -= 1
        elif key == right: col += 1

        log.debug('Keyboard nav on cell {} ' '(new cell: '
                  '{})'.format(self.__selected, (row, col)))

        self.__selectCell(row, col)


    def GetSelection(self):
        """Returns the currently selected item, as a tuple of ``(row, col)``
        indices. If an entire row has been selected, the ``col`` index  will
        be -1, and vice-versa. If nothing is selected, ``None`` is returned.
        """
        return self.__selected


    def SetSelection(self, row, col):
        """Select the given item. A :exc:`ValueError` is raised if the
        selection is invalid.

        :arg row: Row index of item to select. Pass in -1 to select a whole
                  column.

        :arg col: Column index of item to select. Pass in -1 to select a whole
                  row.

        :returns: ``True`` if the selected item was changed, ``False``
                  otherwise.
        """

        if self.__selected == (row, col):
            return False

        nrows, ncols = self.GetGridSize()

        if self.__selectable == 'rows':

            if col != -1 or row < 0 or row >= nrows:
                raise ValueError('Invalid row: {}'.format(row))

        elif self.__selectable == 'columns':

            if row != -1 or col < 0 or col >= ncols:
                raise ValueError('Invalid column: {}'.format(col))

        elif self.__selectable == 'cells':

            if row < 0 or row >= nrows or col < 0 or col >= ncols:
                raise ValueError('Invalid cell: {}, {}'.format(row, col))

        if self.__selected is not None:

            lsrow, lscol = self.__selected
            self.__selected = None
            self.__select(lsrow, lscol, self.__selectable, False)

        self.__selected = row, col

        self.__select(row, col, self.__selectable, True)
        self.__scrollTo(row, col)

        return True


    def __scrollTo(self, row, col):
        """If scrolling is enabled, this method makes sure that the specified
        row/column is visible.
        """

        # No scrolling
        if not (self.__hscroll or self.__vscroll):
            return

        if row == -1: row = 0
        if col == -1: col = 0

        # We're assuming that the
        # scroll rate is in pixels
        startx,    starty    = self                    .GetViewStart()
        sizex,     sizey     = self                    .GetClientSize()
        posx,      posy      = self.__widgets[row][col].GetPosition()
        widgSizex, widgSizey = self.__widgets[row][col].GetSize()

        # Take into account the size
        # of the widget in the cell
        sizex -= widgSizex
        sizey -= widgSizey

        scrollx = startx
        scrolly = starty

        # Figure out if the widget is
        # currently visible and, if
        # not, scroll so that it is.
        if   posx < startx:         scrollx = posx
        elif posx > startx + sizex: scrollx = posx - sizex

        if   posy < starty:         scrolly = posy
        elif posy > starty + sizey: scrolly = posy - sizey

        if scrollx != startx or scrolly != starty:
            self.Scroll(scrollx, scrolly)


    def __select(self, row, col, selectType, select=True):
        """Called by the :meth:`SetSelection` method. Sets the background
        colour of the specified row/column to the selection colour, or the
        default colour.

        :arg row:        Row index. If -1, the colour of the entire column is
                         toggled.

        :arg col:        Column index. If -1, the colour of the entire row is
                         toggled.

        :arg selectType: Either ``'rows'``, ``'columns'``, or ``'cells'``.

        :arg select:     If ``True``, the item background colour is set to the
                         selected colour, otherwise it is set to its default
                         colour.
        """

        nrows, ncols = self.GetGridSize()

        if row == -1 and col == -1:
            return

        if nrows == 0 or ncols == 0:
            return

        if row == -1:
            rows = range(nrows)
            cols = [col] * nrows

        elif col == -1:
            rows = [row] * ncols
            cols = range(ncols)

        else:
            rows = [row]
            cols = [col]

        for row, col in zip(rows, cols):

            if   select:  colour = self.__selectedColour
            elif row % 2: colour = self.__oddColour
            else:         colour = self.__evenColour

            container = self.__widgets[   row][col]
            widget    = self.__widgetRefs[row][col]
            self.__setBackgroundColour(container, colour)
            self.__setBackgroundColour(widget,    colour)
            widget.Refresh()


    def ShowRowLabels(self, show=True):
        """Shows/hides the grid row labels.  The :meth:`Refresh` method must
        be called afterwards for this method to take effect.
        """
        self.__showRowLabels = show


    def ShowColLabels(self, show=True):
        """Shows/hides the grid column labels. The :meth:`Refresh` method must
        be called afterwards for this method to take effect.
        """
        self.__showColLabels = show


    def SetRowLabel(self, row, label):
        """Sets a label for the specified row.

        Raises an :exc:`IndexError` if the row is invalid.
        """
        if row < 0 or row >= self.__nrows:
            raise IndexError('Row {} out of bounds ({})'.format(
                row, self.__nrows))

        self.__rowLabels[row][1].SetLabel(label)


    def SetColLabel(self, col, label):
        """Sets a label for the specified column.

        Raises an :exc:`IndexError` if the column is invalid.
        """
        if col < 0 or col >= self.__ncols:
            raise IndexError('Column {} out of bounds ({})'.format(
                col, self.__ncols))

        self.__colLabels[col][1].SetLabel(label)


WG_SELECTABLE_CELLS = 1
"""If this style is enabled, individual cells can be selected. """


WG_SELECTABLE_ROWS = 2
"""If this style is enabled, whole rows can be selected. """


WG_SELECTABLE_COLUMNS = 4
"""If this style is enabled, whole columns can be selected. """


WG_KEY_NAVIGATION = 8
"""If this style is enabled along with one of the ``*_SELECTABLE_*`` styles,
the user may use the keyboard to navigate between cells, rows, or columns.
"""


_WidgetGridSelectEvent, _EVT_WG_SELECT = wxevent.NewEvent()


EVT_WG_SELECT = _EVT_WG_SELECT
"""Identifier for the :data:`WidgetGridSelectEvent`. """


WidgetGridSelectEvent = _WidgetGridSelectEvent
"""Event generated when an item in a ``WidgetGrid`` is selected. A
``WidgetGridSelectEvent`` has the following attributes:

 - ``row`` Row index of the selected item. If -1, an entire column
   has been selected.

 - ``col`` Column index of the selected item. If -1, an entire row
   has been selected.
"""



#!/usr/bin/env python
#
# progress.py - The Bounce class
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides some classes and functions which use the
``wx.ProgressDialog`` to display the progress of some task.
"""


import threading
import contextlib

import deprecation

import wx

from fsleyes_widgets import isalive


@contextlib.contextmanager
def bounce(*args, **kwargs):
    """Context manager which creates, starts and yields a :class:`Bounce`
    dialog, and destroys it on exit.
    """

    dlg = Bounce(*args, **kwargs)
    dlg.StartBounce()

    try:
        yield dlg
    finally:
        dlg.StopBounce()
        dlg.Destroy()


def runWithBounce(task, *args, **kwargs):
    """Runs the given ``task`` in a separate thread, and creates a
    ``Bounce`` dialog which is displayed while the task is running.

    :arg callback: Must be passed as a keyword argument. A function to call
                   when the ``task`` has finished. Must accept one boolean
                   parameter which is ``True`` if the task ended, or ``False``
                   if the progress dialog was cancelled.

    :arg dlg: Must be passed as a keyword argument. A ``Bounce`` dialog to
                   use. If not provided, one is created. If provided, the
                   caller is responsible for destroying it.

    :arg polltime: Must be passed as a keyword argument. Amount of time in
                   seconds to wait while periodically checking the task
                   state. Defaults to 0.1 seconds.

    All other arguments are passed through to :meth:`Bounce.__init__`,
    unless a ``dlg`` is provided.

    .. note:: This function is non-blocking - it returns immediately. Use
              the ``callback`` function to be notified when the ``task``
              has completed.
    """

    dlg      = kwargs.pop('dlg',      None)
    polltime = kwargs.pop('pollTime', 0.1)
    callback = kwargs.pop('callback', None)
    owndlg   = dlg is None

    if dlg is None:
        dlg = Bounce(*args, **kwargs)

    timer         = wx.Timer(dlg)
    thread        = threading.Thread(target=task)
    thread.daemon = True

    def realCallback(completed):
        dlg.StopBounce()
        timer.Stop()
        if owndlg:
            dlg.Destroy()
        if callback is not None:
            callback(completed)

    def poll(ev):
        if not thread.is_alive():
            realCallback(True)
        elif dlg.WasCancelled():
            realCallback(False)
        else:
            dlg.DoBounce()

    thread.start()
    timer.Start(polltime * 1000, wx.TIMER_CONTINUOUS)
    dlg.Bind(wx.EVT_TIMER, poll)
    dlg.Show()


class Bounce(wx.ProgressDialog):
    """Display a 'bouncing' progress bar.

    The ``Bounce`` class is a ``wx.ProgressDialog`` for use with tasks with an
    unknown duration. The progress bar 'bounces' back and forth until the
    dialog is destroyed or cancelled.

    A ``Bounce`` dialog can either be controlled manually via the
    :meth:`DoBounce` method, , or allowed to run automatically via the
    :meth:`StartBounce`. Automatic bouncing can be stopped via
    :meth:`StopBounce`.
    """


    def __init__(self, title=None, message=None, *args, **kwargs):
        """Create a ``Bounce`` dialog.

        :arg title:   Dialog title.

        :arg message: Dialog message.

        :arg delay:   Must be passed as a keyword argument. Delay in
                      milliseconds between progress bar updates. Defaults to
                      200 milliseconds.

        :arg values:  Must be passed as a keyword argument. A sequence of
                      values from 1 to 99 specifying the locations of the
                      progress bar on each update. Deafults to ``[1, 25, 50,
                      75, 99]``.

        All other arguments are passed through to ``wx.ProgressDialog``.
        """

        if title   is None: title   = 'Title'
        if message is None: message = 'Message'

        self.__delay     = kwargs.pop('delay',  200)
        self.__values    = kwargs.pop('values', [1, 25, 50, 75, 99])
        self.__direction = 1
        self.__index     = 0
        self.__bouncing  = False

        wx.ProgressDialog.__init__(self, title, message, *args, **kwargs)


    @classmethod
    @deprecation.deprecated(deprecated_in='0.3.0',
                            removed_in='1.0.0',
                            details='Use the runWithBounce function instead')
    def runWithBounce(cls, task, *args, **kwargs):
        """Deprecated - use the standalone :func:`runWithBounce` function
        instead.
        """
        return runWithBounce(task, *args, **kwargs)


    def Close(self):
        """Close the ``Bounce`` dialog. """
        self.__bouncing = False
        wx.ProgressDialog.Close(self)


    def EndModal(self, code=wx.ID_OK):
        """Close the ``Bounce`` dialog. """
        self.__bouncing = False
        wx.ProgressDialog.EndModal(self, code)


    def Destroy(self):
        """Destroy the ``Bounce`` dialog. """
        self.__bouncing = False
        wx.ProgressDialog.Destroy(self)


    def StartBounce(self):
        """Start automatic bouncing. """
        self.__bouncing = True
        self.__autoBounce()


    def StopBounce(self):
        """Stop automatic bouncing. """
        self.__bouncing = False


    def Update(self, value, message=None):
        """Overrides ``wx.ProgressDialog.Update``.

        The ``Update`` method in wxPython 3.0.2.0 will raise an error if a
        ``message`` of ``None`` gets passed in. This implementation accepts a
        ``message`` of ``None``.
        """

        if message is None: return super(Bounce, self).Update(value)
        else:               return super(Bounce, self).Update(value, message)


    def UpdateMessage(self, message):
        """Updates the message displayed on the dialog. """
        self.Update(self.__values[self.__index], message)


    def DoBounce(self, message=None):
        """Perform a single bounce update to the progress bar.

        :arg message: New message to show.

        :returns:     ``False`` if the dialog gets cancelled, ``True``
                      otherwise.
        """

        newval = self.__values[self.__index]

        if self.WasCancelled() or \
           not self.Update(newval, message):
            return False

        self.__index += self.__direction

        if self.__index >= len(self.__values):
            self.__index     = len(self.__values) - 2
            self.__direction = -self.__direction
        elif self.__index == 0:
            self.__index     = 0
            self.__direction = -self.__direction

        return True


    def __autoBounce(self):
        """Automatic bouncing.

        If a call to :meth:`StopBounce` has been made, this method does
        nothing.

        Otherwise, calls :meth:`DoBounce` and, if that call returns ``True``,
        schedules a future call to this method.
        """

        # We use a closure, as if this dialog
        # gets destroyed while a call is
        # scheduled, wx segfaults when it tries
        # to call the instance method.
        def realAutoBounce():

            if not isalive(self):
                return

            if not self.__bouncing:
                return

            if self.DoBounce():
                wx.CallLater(self.__delay, realAutoBounce)

        realAutoBounce()
